"""List BaseSpace projects subcommand."""
import backoff

from .utils import get_line
from ....base import Command
from ..... import client


class BaseSpaceList(Command):
    """BaseSpace list command executor."""

    def initialize(self):
        """Initialize basespace projects list subcommand."""
        self.login()

    def validate(self):
        """Need to override because it's not implemented in the base class."""
        pass  # pylint: disable=unnecessary-pass

    @backoff.on_exception(
        backoff.expo,
        (client.APIClientTimeout),
        max_tries=2,
        max_time=30,
    )
    def get_basespace_projects(self, next_link=None):
        """Get BaseSpace projects page."""
        return self.api_client.list_basespace_projects(next_link)

    def get_paginated_basespace_projects(self):
        """Paginate over all BaseSpace projects.

        Yields:
            paginated lists of BaseSpace projects
        """
        more = True
        next_link = None
        while more:
            self.echo_debug("Get BaseSpace projects page")
            resp = self.get_basespace_projects(next_link)

            yield resp.results
            next_link = resp.meta.next
            more = next_link is not None

    def execute(self):
        """Make a request to list BaseSpace projects."""
        self.echo_debug("Listing BaseSpace projects.")

        try:
            for basespace_projects in self.get_paginated_basespace_projects():
                if not basespace_projects:
                    self.echo_debug("No BaseSpace projects were found.")
                    return

                for basespace_project in basespace_projects:
                    self.echo_data(get_line(basespace_project))

        except client.APIClientError as err:
            self.echo_error("There was an error listing BaseSpace projects.")
            if err.status_code == 404:
                self.echo_error("No BaseSpace projects found.")
            raise



"""Gencove CLI models"""
from datetime import datetime
from typing import Any, List, Optional, Union
from uuid import UUID

from pydantic import BaseModel, HttpUrl, validator


# pylint: disable=too-few-public-methods
class GencoveBaseModel(BaseModel):
    """Gencove Base Model"""

    id: UUID


class AccessJWT(BaseModel):
    """AccessJWT model"""

    access: str


class CreateJWT(AccessJWT):
    """CreateJWT model"""

    refresh: str


class S3Object(BaseModel):
    """S3Object model"""

    bucket: Optional[str]
    object_name: Optional[str]


class GencoveStatus(GencoveBaseModel):
    """GencoveStatus model"""

    status: Optional[str]
    note: Optional[str]
    created: Optional[datetime]
    transition_cutoff: Optional[datetime]


class UploadsPostData(GencoveBaseModel):
    """UploadsPostData model"""

    destination_path: Optional[str]
    s3: Optional[S3Object]
    last_status: Optional[GencoveStatus]


class ResponseMeta(BaseModel):
    """ResponseMeta model"""

    count: Optional[int]
    next: Optional[str]
    previous: Optional[str]


# pylint: disable=too-few-public-methods
class PipelineCapabilities(GencoveBaseModel):
    """Pipeline Capabilities record"""

    id: UUID
    name: Optional[str]
    key: Optional[str]


class SampleFile(GencoveBaseModel):
    """SampleFile model"""

    s3_path: Optional[str]
    size: Optional[int]
    download_url: Optional[HttpUrl]
    file_type: Optional[str]
    checksum_sha256: Optional[str]


# pylint: disable=too-few-public-methods
class Project(GencoveBaseModel):
    """Project record"""

    name: Optional[str]
    description: Optional[str]
    created: Optional[datetime]
    organization: Optional[str]
    sample_count: Optional[int]
    pipeline_capabilities: Optional[Union[UUID, PipelineCapabilities]]
    roles: Optional[dict]
    webhook_url: Optional[Union[HttpUrl, str]]  # deprecated
    files: Optional[List[SampleFile]]


class UploadCredentials(BaseModel):
    """UploadCredentials model"""

    version: Optional[int]
    access_key: Optional[str]
    secret_key: Optional[str]
    token: Optional[str]
    expiry_time: Optional[str]  # needs to be str for boto3 to work


class SampleDetails(GencoveBaseModel):
    """SampleDetails model"""

    created: Optional[datetime]
    modified: Optional[datetime]
    run: Optional[str]
    client_id: Optional[str]
    physical_id: Optional[str]
    legacy_id: Optional[str]
    last_status: Optional[GencoveStatus]
    archive_last_status: Optional[GencoveStatus]
    files: Optional[List[SampleFile]]


class ProjectSamples(BaseModel):
    """ProjectSamples model"""

    meta: ResponseMeta
    results: Optional[List[SampleDetails]]


class Upload(BaseModel):
    """Upload model"""

    upload: Optional[UUID]
    destination_path: Optional[str]
    last_status: Optional[GencoveStatus]


class Fastqs(BaseModel):
    """Fastqs model"""

    r1: Optional[Upload]
    r2: Optional[Upload]


class Sample(BaseModel):
    """Sample model"""

    client_id: Optional[str]
    fastq: Optional[Fastqs]
    sample: Optional[UUID]


class SampleSheet(BaseModel):
    """SampleSheet model"""

    meta: ResponseMeta
    results: Optional[List[Sample]]


class UploadSamples(BaseModel):
    """UploadSamples model"""

    uploads: Optional[List[Sample]]
    metadata: Optional[Any]


class QualityControlType(BaseModel):
    """QualityControlType model"""

    key: Optional[str]
    type: Optional[str]


class QualityControlData(BaseModel):
    """QualityControlData model"""

    value_expected: Optional[float]
    value_measured: Optional[float]
    value_string: Optional[str]
    status: Optional[str]

    @validator("value_string", pre=True)
    def blank_string(
        cls, value: str  # noqa: N805
    ):  # pylint: disable=no-self-argument,no-self-use
        """Validator for value_string field, return None in case of empty string"""
        if value == "":
            return None
        return value


class QualityControl(BaseModel):
    """QualityControl model"""

    quality_control_type: Optional[QualityControlType]
    quality_control: Optional[QualityControlData]


class SampleQC(BaseModel):
    """SampleQC model"""

    meta: ResponseMeta
    results: Optional[List[QualityControl]]


class ClientFastQ(BaseModel):
    """ClientFastQ model"""

    client_id: Optional[str]
    fastq: Optional[Upload]


class UploadFastQ(BaseModel):
    """UploadFastQ model"""

    meta: ResponseMeta
    results: Optional[List[ClientFastQ]]


class Projects(BaseModel):
    """Projects model"""

    meta: ResponseMeta
    results: Optional[List[Project]]


class BatchType(BaseModel):
    """BatchType model"""

    key: Optional[str]
    description: Optional[str]


class ProjectBatchTypes(BaseModel):
    """ProjectBatchTypes model"""

    meta: ResponseMeta
    results: Optional[List[BatchType]]


class BatchDetail(GencoveBaseModel):
    """BatchDetail model"""

    name: Optional[str]
    batch_type: Optional[str]
    sample_ids: Optional[List[UUID]]
    last_status: Optional[GencoveStatus]
    files: Optional[List[SampleFile]]


class ProjectBatches(BaseModel):
    """ProjectBatches model"""

    meta: ResponseMeta
    results: Optional[List[BatchDetail]]


class ProjectMergeVCFs(GencoveBaseModel):
    """ProjectMergeVCFs model"""

    created: Optional[datetime]
    user: Optional[UUID]
    last_status: Optional[GencoveStatus]
    up_to_date: Optional[bool]


class SampleMetadata(BaseModel):
    """SampleMetadata model"""

    metadata: Optional[Any]


class BaseSpaceProjectDetail(BaseModel):
    """BaseSpace project detail model"""

    basespace_id: Optional[str]
    basespace_name: Optional[str]
    basespace_date_created: Optional[datetime]


class BaseSpaceProject(BaseModel):
    """BaseSpace project model"""

    meta: ResponseMeta
    results: Optional[List[BaseSpaceProjectDetail]]


class BaseSpaceBiosampleDetail(BaseModel):
    """BaseSpace Biosample detail model"""

    basespace_id: Optional[str]
    basespace_bio_sample_name: Optional[str]
    basespace_date_created: Optional[datetime]


class BaseSpaceBiosample(BaseModel):
    """BaseSpace Biosample model"""

    meta: ResponseMeta
    results: Optional[List[BaseSpaceBiosampleDetail]]


class BaseSpaceProjectImportDetail(GencoveBaseModel):
    """BaseSpace project import detail model"""

    project_id: UUID
    identifier: str
    metadata: Optional[Any]


class BaseSpaceProjectImport(BaseModel):
    """BaseSpace project import model"""

    meta: ResponseMeta
    results: Optional[List[BaseSpaceProjectImportDetail]]


class S3AutoimportTopic(GencoveBaseModel):
    """S3 autoimport topic"""

    topic_arn: str


class S3ProjectImportDetail(GencoveBaseModel):
    """S3 project import detail model"""

    project_id: UUID
    s3_uri: str
    metadata: Optional[Any]


class S3ProjectImport(BaseModel):
    """S3 project import model"""

    meta: ResponseMeta
    results: Optional[List[S3ProjectImportDetail]]


class SampleImport(BaseModel):
    """Existing sample import model"""

    sample_id: UUID
    client_id: Optional[str]


class ImportExistingSamplesModel(BaseModel):
    """Import existing samples model"""

    project_id: UUID
    samples: List[SampleImport]
    metadata: Optional[Any]


class FileType(BaseModel):
    """FileType model"""

    key: Optional[str]
    description: Optional[str]


class FileTypesModel(BaseModel):
    """File types model"""

    meta: ResponseMeta
    results: Optional[List[FileType]]


class Pipeline(GencoveBaseModel):
    """Pipeline model"""

    version: str


class PipelineDetail(Pipeline):
    """Pipeline detail model"""

    capabilities: List[PipelineCapabilities]


class Pipelines(BaseModel):
    """Pipeline model"""

    meta: ResponseMeta
    results: Optional[List[Pipeline]]


class UploadURLImport(GencoveBaseModel):
    """URL import moodel"""

    s3: Optional[S3Object]
    last_status: Optional[GencoveStatus]
    destination_path: Optional[str]
    source_url: Optional[str]


class File(GencoveBaseModel):
    """File model"""

    s3_path: Optional[str]
    size: Optional[int]
    download_url: Optional[str]
    file_type: Optional[str]
    checksum_sha256: Optional[str]


class SampleManifest(GencoveBaseModel):
    """Sample manifest model"""

    file_name: str
    file: Optional[File]
    project: UUID


class SampleManifests(BaseModel):
    """Sample manifests list model"""

    results: List[SampleManifest]



"""VCR filters for samples tests."""
from gencove.tests.filters import _replace_uuid_from_url
from gencove.tests.utils import MOCK_CHECKSUM


def filter_samples_request(request):
    """Filter samples sensitive data from request."""
    return _replace_uuid_from_url(request, "samples")


def filter_files_response_filename(response):
    """Filter file sensitive data from response, using the same filename
    provided in the response.
    """
    if ".sha256" in response["headers"].get("Content-Disposition", [""])[0]:
        filename = response["body"]["string"].decode().split(" *")[1].strip()
        response["body"]["string"] = f"{MOCK_CHECKSUM} *{filename}\n".encode()
    return response



"""Test project's create sample manifest command."""

# pylint: disable=wrong-import-order, import-error
import csv
import operator
import tempfile
import uuid

from click.testing import CliRunner

from gencove.client import APIClient  # noqa: I100
from gencove.command.projects.cli import create_sample_manifest
from gencove.tests.decorators import assert_authorization, assert_no_requests
from gencove.tests.filters import filter_jwt, replace_gencove_url_vcr
from gencove.tests.projects.vcr.filters import (
    filter_project_sample_manifest_request,
)
from gencove.tests.upload.vcr.filters import filter_volatile_dates
from gencove.tests.utils import get_vcr_response

import pytest

from vcr import VCR


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/projects/vcr",
        "filter_headers": [
            "Authorization",
            "Content-Length",
            "User-Agent",
        ],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
            filter_project_sample_manifest_request,
        ],
        "before_record_response": [
            filter_jwt,
            filter_volatile_dates,
        ],
    }


@pytest.fixture(scope="function")
def dummy_valid_manifest_csv():
    """Valid manifest CSV file"""

    manifest_columns = [
        "Plate ID",
        "Well",
        "Unique sample identifier",
        "Buffer",
        "Concentration if known (ng/ul)",
        "Species",
        "Coverage",
    ]

    well_pattern_96 = [
        letter + str(num) for num in range(1, 13) for letter in "ABCDEFGH"
    ]
    with tempfile.NamedTemporaryFile(
        mode="w+", suffix=".csv", delete=False
    ) as tmp_file:
        writer = csv.writer(tmp_file)
        writer.writerow(manifest_columns)  # header
        writer.writerow(["", "A1", "client-id", "", "", "", ""])
        for well in well_pattern_96[1:]:
            writer.writerow(["", well, "", "", "", "", ""])
        tmp_file.seek(0)
    return tmp_file.name


@pytest.mark.vcr
@assert_authorization
def test_create_sample_manifest__success(
    credentials,
    mocker,
    project_id_sample_manifest,
    dummy_valid_manifest_csv,
    recording,
    vcr,
):  # pylint: disable=too-many-arguments,redefined-outer-name
    """Create sample manifest success case"""
    runner = CliRunner()
    if not recording:
        # Mock create_project only if using the cassettes, since we
        # mock the return value.
        create_sample_manifest_response = get_vcr_response(
            "/api/v2/project-sample-manifests/", vcr, operator.contains
        )
        mocked_create_sample_manifest = mocker.patch.object(
            APIClient,
            "create_sample_manifest",
            return_value=create_sample_manifest_response,
        )
    res = runner.invoke(
        create_sample_manifest,
        [
            project_id_sample_manifest,
            dummy_valid_manifest_csv,
            *credentials,
        ],
    )
    assert res.exit_code == 0
    if not recording:
        mocked_create_sample_manifest.assert_called_once()


@pytest.mark.vcr
@assert_authorization
def test_create_sample_manifest__not_owned_project(
    credentials,
    mocker,
    dummy_valid_manifest_csv,
    recording,
):  # pylint: disable=unused-argument,redefined-outer-name
    """Test project not owned case"""
    runner = CliRunner()
    res = runner.invoke(
        create_sample_manifest,
        [
            str(uuid.uuid4()),
            dummy_valid_manifest_csv,
            *credentials,
        ],
    )
    assert (
        "Project does not exist or you do not have privileges to access it"
        in res.output
    )


@assert_no_requests
def test_create_sample_manifest__bad_project_id(
    credentials, mocker, dummy_valid_manifest_csv
):  # pylint: disable=redefined-outer-name
    """Test manifest creation failure when non-uuid string is used as project
    id.
    """
    runner = CliRunner()
    mocked_create_sample_manifest = mocker.patch.object(
        APIClient,
        "create_sample_manifest",
    )
    res = runner.invoke(
        create_sample_manifest,
        [
            "1111111",
            dummy_valid_manifest_csv,
            *credentials,
        ],
    )
    assert res.exit_code == 1
    mocked_create_sample_manifest.assert_not_called()
    assert "Project ID is not valid" in res.output



"""Project restore samples shell command definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.command.utils import validate_uuid, validate_uuid_list
from gencove.constants import Credentials, Optionals
from gencove.logger import echo_debug

from .main import RestoreSamples


@click.command("restore-samples")
@click.argument("project_id", callback=validate_uuid)
@click.option(
    "--sample-ids",
    default="",
    help=(
        "A comma separated list of sample ids which will be restored. If "
        "omitted, restores all archived samples in project."
    ),
    callback=validate_uuid_list,
)
@add_options(common_options)
def restore_project_samples(  # pylint: disable=too-many-arguments
    project_id,
    sample_ids,
    host,
    email,
    password,
    api_key,
):
    """Restore samples in a project.

    `PROJECT_ID`: Gencove project ID
    """
    echo_debug(f"Sample ids translation: {sample_ids}")

    RestoreSamples(
        project_id,
        sample_ids,
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
    ).run()



"""Request project's samples restore."""
from gencove import client  # noqa: I100
from gencove.command.base import Command


class RestoreSamples(Command):
    """Restore project's samples."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        project_id,
        sample_ids,
        credentials,
        options,
    ):
        super().__init__(credentials, options)
        self.project_id = project_id
        self.sample_ids = sample_ids

    def initialize(self):
        """Initialize list subcommand."""
        self.login()

    def validate(self):
        """Validate command input."""

    # no retry for timeouts in order to avoid duplicate heavy operations on
    # the backend
    def execute(self):
        """Make a request to request samples restore for given project."""
        self.echo_debug(
            "Requesting samples restore in project "
            f"{self.project_id} for samples {self.sample_ids}"
        )

        try:
            restored_project_samples_details = self.api_client.restore_project_samples(
                project_id=self.project_id,
                sample_ids=self.sample_ids,
            )
            self.echo_debug(restored_project_samples_details)
            self.echo_info("Request to restore samples accepted.")
        except client.APIClientError as err:
            self.echo_debug(err)
            if err.status_code == 400:
                self.echo_warning(
                    "There was an error requesting project samples restore."
                )
                self.echo_info("The following error was returned:")
                self.echo_info(err.message)
            elif err.status_code == 404:
                self.echo_warning(f"Project {self.project_id} does not exist.")
            else:
                raise



"""Commands to be executed from command line."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials

from .constants import UploadOptions
from .main import Upload


@click.command()
@click.argument("source")
@click.argument("destination", required=False)
@add_options(common_options)
@click.option(
    "--run-project-id",
    default=None,
    help="Immediately assign all uploaded files to this project and run them",
)
@click.option(
    "--output",
    required=False,
    default=None,
    help=(
        "A destination where to store the resulting assignments. "
        "Only compatible with --run-project-id."
    ),
)
@click.option(
    "--no-progress",
    is_flag=True,
    help="If specified, no progress bar is shown.",
)
@click.option(
    "--metadata",
    required=False,
    default=None,
    help=(
        "Assign JSON metadata to all samples created from uploads. "
        "Only compatible with --run-project-id."
    ),
)
def upload(  # pylint: disable=E0012,C0330,R0913
    source,
    destination,
    host,
    email,
    password,
    api_key,
    run_project_id,
    output,
    no_progress,
    metadata,
):  # noqa: D301
    """Upload FASTQ files to Gencove's system.

    `SOURCE`: folder that contains fastq files to be uploaded (acceptable file
    extensions are .fastq.gz, .fastq.bgz, .fq.gz, .fq.bgz), OR .fastq-map.csv
    file with the mapping of R1/R2 files to related sample by client_id

    `DESTINATION` (optional): `gncv://[folder]`, where the folder is the location
    on Gencove systems

    Examples:

        Upload directory contents:

            gencove upload test_dataset gncv://test
    """
    Upload(
        source,
        destination,
        Credentials(email=email, password=password, api_key=api_key),
        UploadOptions(host=host, project_id=run_project_id, metadata=metadata),
        output,
        no_progress,
    ).run()



"""Commands to be executed from command line."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials, Optionals

from .main import ListFileTypes


@click.command("file-types")
@click.option("--project-id", required=False, help="Gencove project ID", type=str)
@add_options(common_options)
def list_file_types(project_id, host, email, password, api_key):
    """List file types in Gencove's system."""
    ListFileTypes(
        project_id,
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
    ).run()



"""List S3 AutoImport jobs subcommand."""
import backoff

from .utils import get_line
from ....base import Command
from ..... import client


class S3AutoImportList(Command):
    """S3 autoimport list command executor."""

    def initialize(self):
        """Initialize s3-autoimport subcommand."""
        self.login()

    def validate(self):
        """Validate command input."""

    def execute(self):
        """Make a request to list jobs to periodically import from S3 URI."""
        self.echo_debug("List AutoImport jobs.")

        try:
            for s3_autoimports in self.get_paginated_s3_autoimports():
                if not s3_autoimports:
                    self.echo_debug("No S3 autoimport jobs were found.")
                    return
                for s3_autoimport in s3_autoimports:
                    self.echo_data(get_line(s3_autoimport))
        except client.APIClientError:
            self.echo_error("There was an error listing S3 autoimport jobs.")
            raise

    def get_paginated_s3_autoimports(self):
        """Paginate over all S3 autoimports.

        Yields:
            paginated lists of S3 autoimports
        """
        more = True
        next_link = None
        while more:
            self.echo_debug("Get S3 autoimport page")
            resp = self.get_s3_autoimports(next_link)

            yield resp.results
            next_link = resp.meta.next
            more = next_link is not None

    @backoff.on_exception(
        backoff.expo,
        (client.APIClientTimeout),
        max_tries=2,
        max_time=30,
    )
    def get_s3_autoimports(self, next_link=None):
        """Get S3 autoimports page."""
        return self.api_client.list_s3_autoimport_jobs(next_link)



"""Tests for utils of Gencove CLI."""
# pylint: disable=wrong-import-order, import-error

import csv
import os
from enum import Enum

from click import Abort
from click.core import Argument, Option
from click.testing import CliRunner

from faker import Faker

from gencove.client import APIClient, APIClientError
from gencove.command.download.utils import (
    _get_prefix_parts,
    build_file_path,
    get_download_template_format_params,
)
from gencove.command.upload.utils import (
    _validate_header,
    _validate_parsed_map,
    looks_like_url,
    parse_fastqs_map_file,
    upload_file,
    valid_fastq_file_name_in_url,
)
from gencove.command.utils import is_valid_uuid, validate_uuid, validate_uuid_list
from gencove.constants import (
    ApiEndpoints,
    Credentials,
    DOWNLOAD_TEMPLATE,
    DownloadTemplateParts,
)
from gencove.exceptions import MaintenanceError, ValidationError
from gencove.utils import enum_as_dict, login

import pytest

fake = Faker()


def test_upload_file(mocker):
    """Sanity check upload function."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        with open("foo.txt", "w", encoding="utf-8") as fastq_file:
            fastq_file.write("AAABBB")
        mocked_s3_client = mocker.Mock()
        assert upload_file(
            mocked_s3_client,
            "foo.txt",
            "foo-bucket",
            object_name="foo-object.txt",
        )
        mocked_s3_client.upload_file.assert_called_once()


def test_download_template_tokens():
    """Ensure download tokens are only the ones defined."""
    assert [
        "client_id",
        "gencove_id",
        "file_type",
        "file_extension",
        "default_filename",
    ] == list(enum_as_dict(DownloadTemplateParts).values())


# pylint: disable=too-many-locals
def test_build_file_path():
    """Test token combinations when building a file path."""
    client_id = "12345"
    gencove_id = "1"
    deliverable = {
        "download_url": "https://example.com/file.txt",
        "file_type": "txt",
    }
    download_template_format_params = get_download_template_format_params(
        client_id, gencove_id
    )
    # file with prefix will never have client_id and gencove_id. 0 is default
    file_with_prefix0 = (
        f"{download_template_format_params['client_id']}/"
        f"{download_template_format_params['gencove_id']}/"
        f"{download_template_format_params['default_filename']}"
    )
    file_with_prefix1 = download_template_format_params["client_id"]
    file_with_prefix2 = download_template_format_params["gencove_id"]
    file_with_prefix3 = "{file_type}"
    file_with_prefix4 = "{file_extension}"
    file_with_prefix5 = "{default_filename}"
    file_with_prefix6 = "{default_filename}.{file_extension}"
    file_with_prefix7 = (
        f"{download_template_format_params['default_filename']}."
        f"{download_template_format_params['client_id']}"
    )
    file_with_prefix8 = (
        f"{download_template_format_params['client_id']}."
        f"{download_template_format_params['default_filename']}"
    )
    file_with_prefix9 = (
        f"{download_template_format_params['file_extension']}_"
        f"{download_template_format_params['client_id']}"
    )
    download_to = "."
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = build_file_path(deliverable, file_with_prefix0, download_to)
        # default_filename is always a full filename with any extensions
        assert result == (
            f"./{download_template_format_params['client_id']}/"
            f"{download_template_format_params['gencove_id']}/file.txt"
        )
        result = build_file_path(deliverable, file_with_prefix1, download_to)
        assert result == f"./{download_template_format_params['client_id']}"
        result = build_file_path(deliverable, file_with_prefix2, download_to)
        assert result == f"./{download_template_format_params['gencove_id']}"
        result = build_file_path(deliverable, file_with_prefix3, download_to)
        assert result == "./txt"
        result = build_file_path(deliverable, file_with_prefix4, download_to)
        assert result == "./txt"
        result = build_file_path(deliverable, file_with_prefix5, download_to)
        assert result == "./file.txt"
        result = build_file_path(deliverable, file_with_prefix6, download_to)
        assert result == "./file.txt.txt"
        result = build_file_path(deliverable, file_with_prefix7, download_to)
        assert result == f"./file.txt.{download_template_format_params['client_id']}"
        result = build_file_path(deliverable, file_with_prefix8, download_to)
        assert result == f"./{download_template_format_params['client_id']}.file.txt"
        result = build_file_path(deliverable, file_with_prefix9, download_to)
        assert result == f"./txt_{download_template_format_params['client_id']}"


def test___get_filename_dirs_prefix():
    """Test proper processing of parts in download template."""
    client_id = "12345"
    gencove_id = "1"
    download_template_format_params = get_download_template_format_params(
        client_id, gencove_id
    )
    template = DOWNLOAD_TEMPLATE.format(**download_template_format_params)

    resp = _get_prefix_parts(template)

    assert resp.dirs == (
        f"{download_template_format_params['client_id']}/"
        f"{download_template_format_params['gencove_id']}"
    )
    assert resp.filename == f"{{{DownloadTemplateParts.DEFAULT_FILENAME.value}}}"
    assert resp.file_extension == ""

    template2 = (
        f"{download_template_format_params['client_id']}-"
        f"{download_template_format_params['gencove_id']}_"
        f"{download_template_format_params['file_type']}"
    )
    resp = _get_prefix_parts(template2)
    assert resp.dirs == ""
    assert resp.filename == (
        f"{download_template_format_params['client_id']}-"
        f"{download_template_format_params['gencove_id']}_"
        f"{{{DownloadTemplateParts.FILE_TYPE.value}}}"
    )
    assert resp.file_extension == ""

    template3 = (
        f"{download_template_format_params['client_id']}-"
        f"{download_template_format_params['gencove_id']}_"
        f"{download_template_format_params['file_type']}.vcf.gz"
    )
    resp = _get_prefix_parts(template3)
    assert resp.dirs == ""
    assert resp.filename == (
        f"{download_template_format_params['client_id']}-"
        f"{download_template_format_params['gencove_id']}_"
        f"{{{DownloadTemplateParts.FILE_TYPE.value}}}"
    )
    assert resp.file_extension == "vcf.gz"


def test_parse_fastqs_map_file():
    """Test parsing of map file into dict."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("test_dir")
        with open("test_dir/test.fastq.gz", "w", encoding="utf-8") as fastq_file1:
            fastq_file1.write("AAABBB")

        with open("test_dir/test_2.fastq.gz", "w", encoding="utf-8") as fastq_file2:
            fastq_file2.write("AAABBB")

        with open("test_dir/test_3.fastq.gz", "w", encoding="utf-8") as fastq_file3:
            fastq_file3.write("AAABBB")

        with open("test_map.csv", "w", encoding="utf-8") as map_file:
            writer = csv.writer(map_file)
            writer.writerows(
                [
                    ["client_id", "r_notation", "path"],
                    ["barid", "r1", "test_dir/test.fastq.gz"],
                    ["barid", "r2", "test_dir/test_2.fastq.gz"],
                    ["barid", "r1", "test_dir/test_3.fastq.gz"],
                ]
            )

        fastqs = parse_fastqs_map_file("test_map.csv")
        assert len(fastqs) == 2
        assert ("barid", "R1") in fastqs
        assert ("barid", "R2") in fastqs
        assert len(fastqs[("barid", "R1")]) == 2
        assert fastqs[("barid", "R1")][0] == "test_dir/test.fastq.gz"
        assert fastqs[("barid", "R1")][1] == "test_dir/test_3.fastq.gz"


def test_invalid_fastqs_map_file():
    """Test that error is raised if file is invalid."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("test_dir")
        with open("test_dir/test.fastq.gz", "w", encoding="utf-8") as fastq_file1:
            fastq_file1.write("AAABBB")

        with open("test_dir/test_2.fastq.gz", "w", encoding="utf-8") as fastq_file2:
            fastq_file2.write("AAABBB")

        with open("test_dir/test_3.fastq.gz", "w", encoding="utf-8") as fastq_file3:
            fastq_file3.write("AAABBB")

        with open("test_map.csv", "w", encoding="utf-8") as map_file:
            writer = csv.writer(map_file)
            writer.writerows(
                [
                    ["client_id", "r_notation", "path"],
                    ["barid", "r1", "test_dir/test.fastq.gz"],
                    ["barid", "r2", "test_dir/test_2.fastq.zip"],
                    ["barid", "r1", "test_dir/test_3.fastq.gz"],
                ]
            )
        try:
            parse_fastqs_map_file("test_map.csv")
        except ValidationError as err:
            assert "Bad file extension in path" in err.args[0]


def test_fastqs_map_file_path_does_not_exist():
    """Test that error is raised if file is invalid."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("test_dir")
        with open("test_dir/test.fastq.gz", "w", encoding="utf-8") as fastq_file1:
            fastq_file1.write("AAABBB")

        with open("test_dir/test_2.fastq.gz", "w", encoding="utf-8") as fastq_file2:
            fastq_file2.write("AAABBB")

        with open("test_dir/test_3.fastq.gz", "w", encoding="utf-8") as fastq_file3:
            fastq_file3.write("AAABBB")

        with open("test_map.csv", "w", encoding="utf-8") as map_file:
            writer = csv.writer(map_file)
            writer.writerows(
                [
                    ["client_id", "r_notation", "path"],
                    ["barid", "r1", "test_dir/test.fastq.gz"],
                    ["barid", "r1", "test_dir/test_3.fastq.gz"],
                    ["barid", "r1", "test_dir/i_dont_exist.fastq.gz"],
                ]
            )
        try:
            parse_fastqs_map_file("test_map.csv")
        except ValidationError as err:
            assert "Could not find" in err.args[0]


def test_fastqs_map_file_has_wrong_header():
    """Test that header is validated properly."""
    runner = CliRunner()
    # This context manager is used to avoid unnecessarily creating a
    # test_map.csv file in the root of the project on every test run
    with runner.isolated_filesystem():
        with open("test_map.csv", "w", encoding="utf-8") as map_file:
            writer = csv.writer(map_file)
            writer.writerows(
                [
                    ["client_id", "r1_notation", "path"],
                    ["barid", "r1", "test_dir/test.fastq.gz"],
                    ["barid", "r1", "test_dir/test_3.fastq.gz"],
                    ["barid", "r1", "test_dir/i_dont_exist.fastq.gz"],
                ]
            )
        try:
            parse_fastqs_map_file("test_map.csv")
        except ValidationError as err:
            assert "Unexpected CSV header" in err.args[0]


def test_parse_fastqs_map_file_with_urls():
    """Test parsing of map file with URLs into dict"""
    runner = CliRunner()
    with runner.isolated_filesystem():
        with open("test_url_map.fastq-map.csv", "w", encoding="utf-8") as map_file:
            writer = csv.writer(map_file)
            writer.writerows(
                [
                    ["client_id", "r_notation", "path"],
                    ["barid", "r1", "https://s3.amazonaws.com/samples/1/R1.fastq.gz"],
                    ["barid", "r2", "https://s3.amazonaws.com/samples/1/R2.fastq.gz"],
                    [
                        "barid",
                        "r1",
                        "https://storage.googleapis.com/samples/2/R1.fastq.gz",
                    ],
                ]
            )

        fastqs = parse_fastqs_map_file("test_url_map.fastq-map.csv")
        assert len(fastqs) == 2
        assert ("barid", "R1") in fastqs
        assert ("barid", "R2") in fastqs
        assert len(fastqs[("barid", "R1")]) == 2
        assert (
            fastqs[("barid", "R1")][0]
            == "https://s3.amazonaws.com/samples/1/R1.fastq.gz"
        )
        assert (
            fastqs[("barid", "R1")][1]
            == "https://storage.googleapis.com/samples/2/R1.fastq.gz"
        )


def test_parse_fastqs_map_file_with_urls_invalid():
    """Test parsing of map file with URLs into dict, invalid URLs"""
    runner = CliRunner()
    with runner.isolated_filesystem():
        with open("test_url_map.fastq-map.csv", "w", encoding="utf-8") as map_file:
            writer = csv.writer(map_file)
            writer.writerows(
                [
                    ["client_id", "r_notation", "path"],
                    ["barid", "r1", "https://s3.amazonaws.com/samples/1/invalid"],
                    ["barid", "r2", "https://s3.amazonaws.com/samples/1/invalid"],
                    [
                        "barid",
                        "r1",
                        "https://storage.googleapis.com/samples/2/invalid",
                    ],
                ]
            )
        try:
            parse_fastqs_map_file("test_url_map.fastq-map.csv")
        except ValidationError as err:
            assert "Could not determine FASTQ file" in err.args[0]


def test__validate_header():
    """Test that header is validated properly."""
    header_row = dict(foo="foo", bar="bar")
    try:
        _validate_header(header_row)
    except ValidationError:
        pass

    # expected
    header_row = dict(client_id="client_id", r_notation="r_notation", path="path")
    assert _validate_header(header_row) is None

    # with spacing
    header_row = dict(client_id="client_id", r_notation=" r_notation", path="path  ")
    assert _validate_header(header_row) is None

    # with uppercase
    header_row = dict(client_id="client_id", r_notation="R_notation", path="path")
    assert _validate_header(header_row) is None


def test_validate_parsed_map__all_urls():
    """Test validation of parsed map, all URLs OK"""
    runner = CliRunner()
    with runner.isolated_filesystem():
        with open("test_url_map.fastq-map.csv", "w", encoding="utf-8") as map_file:
            writer = csv.writer(map_file)
            writer.writerows(
                [
                    ["client_id", "r_notation", "path"],
                    ["url", "r1", "https://s3.amazonaws.com/samples/1/R1.fastq.gz"],
                    ["url", "r2", "https://s3.amazonaws.com/samples/1/R2.fastq.gz"],
                ]
            )
        fastqs = parse_fastqs_map_file("test_url_map.fastq-map.csv")
    assert _validate_parsed_map(fastqs) is None


def test_validate_parsed_map__all_paths():
    """Test validation of parsed map, all paths OK"""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("test_dir")
        with open("test_dir/file_r1.fastq.gz", "w", encoding="utf-8") as fastq_file1:
            fastq_file1.write("AAABBB")

        with open("test_dir/file_r2.fastq.gz", "w", encoding="utf-8") as fastq_file2:
            fastq_file2.write("AAABBB")

        with open("test_file_map.fastq-map.csv", "w", encoding="utf-8") as map_file:
            writer = csv.writer(map_file)
            writer.writerows(
                [
                    ["client_id", "r_notation", "path"],
                    ["file", "r1", "test_dir/file_r1.fastq.gz"],
                    ["file", "r2", "test_dir/file_r2.fastq.gz"],
                ]
            )
        fastqs = parse_fastqs_map_file("test_file_map.fastq-map.csv")
    assert _validate_parsed_map(fastqs) is None


def test_validate_parsed_map__mixed():
    """Test validation of parsed map to prevent heterogeneous path values"""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("test_dir")
        with open("test_dir/file_r1.fastq.gz", "w", encoding="utf-8") as fastq_file1:
            fastq_file1.write("AAABBB")

        with open("test_dir/file_r2.fastq.gz", "w", encoding="utf-8") as fastq_file2:
            fastq_file2.write("AAABBB")

        with open("test_mixed_map.fastq-map.csv", "w", encoding="utf-8") as map_file:
            writer = csv.writer(map_file)
            writer.writerows(
                [
                    ["client_id", "r_notation", "path"],
                    ["url", "r1", "https://s3.amazonaws.com/samples/1/R1.fastq.gz"],
                    ["url", "r2", "https://s3.amazonaws.com/samples/1/R1.fastq.gz"],
                    ["file", "r1", "test_dir/file_r1.fastq.gz"],
                    ["file", "r2", "test_dir/file_r2.fastq.gz"],
                ]
            )
        try:
            parse_fastqs_map_file("test_mixed_map.fastq-map.csv")
        except ValidationError as err:
            assert "Detected both URLs and file paths" in err.args[0]


def test_is_valid_uuid__is_valid():
    """Test that a UUID is a valid UUID"""
    assert is_valid_uuid("11111111-1111-1111-1111-111111111111")


def test_is_valid_uuid__is_not_valid__too_long():
    """Test that UUID with extra chars is not a valid UUID"""
    assert is_valid_uuid("11111111-1111-1111-1111-11111111111122") is False


def test_is_valid_uuid__is_not_valid__too_short():
    """Test that UUID with missing chars is not a valid UUID"""
    assert is_valid_uuid("11111111-1111-1111-1111-1") is False


def test_is_valid_uuid__is_not_valid__text():
    """Test that random word is not a valid UUID"""
    assert is_valid_uuid("foo") is False


def test_enum_as_dict():
    """Test that the function enum_as_dict returns properly formatted dict."""

    class TestEnum(Enum):
        """Enum for testing"""

        KEY1 = "key1"
        KEY2 = "key2"
        KEY3 = "key3"

    assert enum_as_dict(TestEnum) == {
        "KEY1": "key1",
        "KEY2": "key2",
        "KEY3": "key3",
    }


def test_login_mfa(mocker):
    """Test that the the login function in APIClient asks for the one time
    password token.
    """
    api_client = APIClient()
    credentials = Credentials(email="foo@bar.com", password="123456", api_key="")

    def _request(
        endpoint,
        params,
        *args,  # pylint: disable=unused-argument
        **kwargs,  # pylint: disable=unused-argument
    ):
        if endpoint == ApiEndpoints.GET_JWT.value:
            if "otp_token" not in params:
                raise APIClientError(
                    {"otp_token": ["Please enter your OTP token."]}, 401
                )
            return {"access": "access", "refresh": "refresh"}
        return {}

    mocked_request = mocker.patch.object(APIClient, "_request", side_effect=_request)
    mocked_prompt = mocker.patch("gencove.utils.click.prompt", return_value="token")

    login(api_client, credentials)
    assert mocked_request.call_count == 2
    mocked_prompt.assert_called_once_with("One time password", type=str, err=True)


def test_login__returns_maintenance_error_message(mocker):
    """Test that the the login function in APIClient asks for the one time
    password token.
    """
    api_client = APIClient()
    credentials = Credentials(email="foo@bar.com", password="123456", api_key="")

    def _request(
        endpoint,
        params,
        *args,  # pylint: disable=unused-argument
        **kwargs,  # pylint: disable=unused-argument
    ):
        if endpoint == ApiEndpoints.GET_JWT.value:
            raise MaintenanceError(
                {
                    "maintenance": True,
                    "maintenance_eta": "",
                    "maintenance_message": (
                        "Gencove is currently undergoing maintenance and"
                        "will return at the given ETA."
                        "Thank you for your patience."
                    ),
                },
                503,
            )
        return {}

    mocked_request = mocker.patch.object(APIClient, "_request", side_effect=_request)

    with pytest.raises(MaintenanceError):
        login(api_client, credentials)
    mocked_request.assert_called_once()


def test_uuid_without_hyphens_is_converted_to_uuid_with_hyphens():
    """Test that when a uuid without hyphens is passed to the method, it
    will automatically add the hyphens
    """
    expected_uuid = str(fake.uuid4())
    input_uuid_without_hyphens = expected_uuid.replace("-", "")

    param = Argument(["-project_id"])
    actual_uuid = validate_uuid(None, param, input_uuid_without_hyphens)

    assert isinstance(actual_uuid, str)
    assert actual_uuid == expected_uuid


def test_uuid_with_hyphens_remains_as_is():
    """Test that when uuid with hyphens it will not do anything"""
    input_uuid = fake.uuid4()
    expected_uuid = str(input_uuid)

    param = Argument(["-project_id"])
    actual_uuid = validate_uuid(None, param, input_uuid)

    assert isinstance(actual_uuid, str)
    assert actual_uuid == expected_uuid


def test_uuid_with_hyphens_remains_as_is__not_v4_uuid():
    """Test that when uuid with hyphens it will not do anything even for values
    that are not technically a uuid v4.
    """
    input_uuid = "11111111-1111-1111-1111-111111111234"
    expected_uuid = str(input_uuid)

    param = Argument(["-project_id"])
    actual_uuid = validate_uuid(None, param, input_uuid)

    assert isinstance(actual_uuid, str)
    assert actual_uuid == expected_uuid


def test_validate_uuid__raises_if_uuid_invalid():
    """Test that an invalid uuid will raise a click.Abort"""
    input_uuid = "codef00d-1111-abcd-1111"

    param = Argument(["-project_id"])
    with pytest.raises(Abort):
        validate_uuid(None, param, input_uuid)


def test_validate_uuid_list__returns_list_of_valid_uuids():
    """Test string of valid uuids returns list of valid uuids"""
    param = Option(["-sample_ids"], nargs=1, multiple=False, default=None)
    valid_uuid_list = fake.pylist(nb_elements=3, value_types=("uuid4",))
    valid_uuids_string = ",".join(valid_uuid_list)
    validation_result = validate_uuid_list(None, param, valid_uuids_string)

    assert isinstance(validation_result, list)
    assert len(validation_result) == len(valid_uuid_list)


def test_validate_uuid_list__raises_if_not_all_ids_valid():
    """Test uuid list containing one invalid uuid will raise a click.Abort"""
    invalid_uuid = "codef00d-1111-abcd-1111"
    param = Option(["-sample_ids"], nargs=1, multiple=False, default=None)
    valid_uuid_list = fake.pylist(nb_elements=3, value_types=("uuid4",))
    valid_uuids_string = ",".join(valid_uuid_list)
    uuids_string = f"{valid_uuids_string},{invalid_uuid}"

    with pytest.raises(Abort):
        validate_uuid_list(None, param, uuids_string)


@pytest.mark.parametrize(
    "url,expected",
    [
        ("http://example.com/file.fastq.bgz", True),
        ("http://example.com/file.fastq.gz", True),
        ("http://example.com/file.fq.bgz", True),
        ("http://example.com/file.fq.gz", True),
        ("http://example.com/file.txt", False),
        ("http://example.com/file", False),
        ("http://example.com/file.fastq", False),
        ("http://example.com/.fastq", False),
        ("http://example.com/", False),
        ("foo", False),
    ],
)
def test_file_name_in_url(url, expected):
    """Test multiple cases to ensure valid FASTQ file detected"""
    assert valid_fastq_file_name_in_url(url) == expected


@pytest.mark.parametrize(
    "value,expected",
    [
        ("http://example.com", True),
        ("https://example.com", True),
        ("http://example.com/file.fastq", True),
        ("https://www.example.com", True),
        ("http://", False),
        ("https://", False),
        ("example.com", False),
        ("ftp://example.com", False),
        ("foo", False),
        ("", False),
    ],
)
def test_looks_like_url(value, expected):
    """Test multiple cases of URL strings"""
    assert looks_like_url(value) == expected



"""Utilities for processing projects."""
from gencove.command.utils import sanitize_string


def get_line(project, include_capability):
    """Build a list of relevant data to be printed.

    Args:
        project (Project): instance of project

    Returns:
        list(str): list of relevant data to be printed
    """
    output_list = [
        str(project.created),
        str(project.id),
        sanitize_string(project.name),
        sanitize_string(project.pipeline_capabilities.name),
    ]
    if include_capability:
        output_list.append(str(project.pipeline_capabilities.id))
        output_list.append(str(project.pipeline_capabilities.key))

    return "\t".join(output_list)



"""Utilities for processing BaseSpace projects."""
from gencove.command.utils import sanitize_string


def get_line(basespace_project):
    """Build a list of relevant data to be printed.

    Args:
        basespace_project (BaseSpaceProject): instance of BaseSpace project

    Returns:
        list(str): list of relevant data to be printed
    """
    return "\t".join(
        [
            str(basespace_project.basespace_date_created),
            str(basespace_project.basespace_id),
            sanitize_string(basespace_project.basespace_name),
        ]
    )



"""List projects shell command definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials, Optionals

from .main import List


@click.command("list")
@click.option(
    "--include-capability/--no-include-capability",
    help="Include capability id and key",
    is_flag=True,
)
@add_options(common_options)
def list_projects(include_capability, host, email, password, api_key):
    """List your projects."""
    List(
        include_capability,
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
    ).run()



"""Commands to be executed from command line."""
# pylint: disable=E0012,C0330,R0913
import click

from .autoimport_list.cli import autoimport_list
from .create.cli import create


@click.group()
def autoimports():
    """BaseSpace autoimports management commands."""


autoimports.add_command(autoimport_list)
autoimports.add_command(create)



"""Utilities for different levels of click.echo outputs."""
import logging
import os
import sys
import uuid
from datetime import datetime
from platform import platform

import boto3

import click

from gencove.version import version

INFO = "INFO"
DEBUG = "DEBUG"
LOG_LEVEL = os.environ.get("GENCOVE_LOG", INFO)
DEBUG_LOG = []


def _echo(msg, **kwargs):
    """Output click echo message."""
    if LOG_LEVEL == DEBUG:
        msg = f"{datetime.utcnow().isoformat()} {msg}"
    click.echo(msg, **kwargs)


def _log(msg):
    """Adds the message to the log list."""
    msg = f"{datetime.utcnow().isoformat()} {msg}"
    DEBUG_LOG.append(msg)


def output_warning(text):
    """Click echo warning."""
    return click.style(f"WARNING: {text}", fg="bright_yellow", bold=True)


def output_error(text):
    """Click echo error."""
    return click.style(f"ERROR: {text}", fg="bright_red", bold=True)


def echo_data(msg, **kwargs):
    """Output click echo msg."""
    _echo(msg, **kwargs)
    _log(msg)


def echo_info(msg, **kwargs):
    """Output click echo msg."""
    _echo(msg, err=True, **kwargs)
    _log(msg)


def echo_debug(msg, **kwargs):
    """Output click echo msg only if debug is on."""
    if LOG_LEVEL == DEBUG:
        _echo(msg, err=True, **kwargs)
    _log(msg)


def echo_warning(msg, **kwargs):
    """Output click echo msg with background."""
    _echo(output_warning(msg), err=True, **kwargs)
    _log(msg)


def echo_error(msg, **kwargs):
    """Output click echo msg with background."""
    _echo(output_error(msg), err=True, **kwargs)
    _log(msg)


echo_debug(
    f"Python version: {sys.version_info.major}."
    f"{sys.version_info.minor}."
    f"{sys.version_info.micro}"
)
echo_debug(f"CLI version: {version()}")
echo_debug(f"OS details: {platform()}")
echo_debug(f"boto3 version: {boto3.__version__}")
if LOG_LEVEL == DEBUG:
    logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
    logging.getLogger("botocore.auth").setLevel(logging.CRITICAL)


def get_debug_file_name():
    """Creates a 'unique' file name YYYY_MM_DD_HH_MM_RANDOM_ID.log
    inside a YYYY_MM folder.
    """
    now = datetime.utcnow()
    timestamp = f"{now:%Y_%m_%d_%H_%M_%S}"
    random_id = str(uuid.uuid4())[:8]
    filename = f"{timestamp}_{random_id}.log"
    folder_name = f"{now:%Y_%m}"
    folder = os.path.join(".logs", folder_name)
    file_path = os.path.join(folder, filename)
    if not os.path.exists(folder):
        os.makedirs(folder)
    return file_path


def dump_debug_log():
    """Saves the logs into a file."""
    if os.environ.get("GENCOVE_SAVE_DUMP_LOG") != "FALSE":
        try:
            debug_filename = get_debug_file_name()
            log = "\n".join(DEBUG_LOG)
            with open(debug_filename, "w", encoding="utf-8") as dump_file:
                dump_file.write(log)
            _echo(
                f"Please attach the debug log file located in {debug_filename} to a bug report."  # noqa: E501  # pylint: disable=line-too-long
            )
        except PermissionError:
            echo_warning(
                "An error occurred, but we couldn't write a debug log file to disk."  # noqa: E501  # pylint: disable=line-too-long
            )



"""Project create merged VCF shell command definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.command.utils import validate_uuid
from gencove.constants import Credentials, Optionals

from .main import CreateMergedVCF


@click.command("create-merged-vcf")
@click.argument("project_id", callback=validate_uuid)
@add_options(common_options)
def create_merged_vcf(
    project_id,
    host,
    email,
    password,
    api_key,
):
    """Merge VCF files in a project.

    `PROJECT_ID`: Gencove project ID
    """
    CreateMergedVCF(
        project_id,
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
    ).run()



"""Project autoimport Biosamples from BaseSpace projects shell command
definition.
"""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.command.utils import validate_uuid
from gencove.constants import Credentials

from .constants import BaseSpaceAutoImportOptionals
from .main import BaseSpaceAutoImport


@click.command("create")
@click.argument("project_id", callback=validate_uuid)
@click.argument("identifier")
@click.option(
    "--metadata-json",
    required=False,
    default=None,
    help=(
        "Add metadata to all samples that are to be imported from "
        "BaseSpace to a project."
    ),
)
@add_options(common_options)
def create(  # pylint: disable=too-many-arguments
    project_id,
    identifier,
    metadata_json,
    host,
    email,
    password,
    api_key,
):  # pylint: disable=line-too-long
    """Sets up periodic import of BaseSpace projects (their Biosamples) whose name contain the identifer
    to a project in Gencove. Optionally assign metadata to the samples to be added when the automatic
    import job runs.

    `PROJECT_ID`: Gencove project ID

    `IDENTIFIER`: string used for identifying projects on BaseSpace

    Examples:

        Set up automatic import of BaseSpace projects' samples that contain the identifier to a project:

            gencove basespace autoimports create 06a5d04b-526a-4471-83ba-fb54e0941758 identifier-in-basespace-project-name

        Set up automatic import of BaseSpace projects' samples that contain the identifier to a project with metadata:

            gencove basespace autoimports create 06a5d04b-526a-4471-83ba-fb54e0941758 identifier-in-basespace-project-name --metadata-json='{"batch": "batch1"}'
    """  # noqa: E501

    BaseSpaceAutoImport(
        project_id,
        identifier,
        Credentials(email=email, password=password, api_key=api_key),
        BaseSpaceAutoImportOptionals(host=host, metadata_json=metadata_json),
    ).run()



"""Gencove CLI utils."""
import os
import re
import sys

import boto3

from botocore.credentials import RefreshableCredentials
from botocore.session import get_session

import click

import progressbar

from gencove.client import APIClientError  # noqa: I100
from gencove.constants import (
    MINIMUM_SUPPORTED_PYTHON_MAJOR,
    MINIMUM_SUPPORTED_PYTHON_MINOR,
)
from gencove.logger import echo_debug, echo_error, echo_info, echo_warning

KB = 1024
MB = KB * 1024
GB = MB * 1024
NUM_MB_IN_CHUNK = 100
CHUNK_SIZE = NUM_MB_IN_CHUNK * MB
FILENAME_RE = re.compile("filename=(.+)")


def get_s3_client_refreshable(refresh_method):
    """Return thread-safe s3 client with refreshable credentials.

    :param refresh_method: function that can get fresh credentials
    """

    def refresh_to_dict():
        """Turn pydantic model into `dict`. Needed for botocore."""
        return refresh_method().dict()

    session = get_session()
    session_credentials = RefreshableCredentials.create_from_metadata(
        metadata=refresh_to_dict(),
        refresh_using=refresh_to_dict,
        method="sts-assume-role",
    )
    # pylint: disable=protected-access
    session._credentials = session_credentials
    boto3_session = boto3.Session(botocore_session=session)
    return boto3_session.client(
        "s3",
        endpoint_url=os.environ.get("GENCOVE_LOCALSTACK_S3_ENDPOINT") or None,
    )


def get_progress_bar(total_size, action):
    """Get progressbar.ProgressBar instance for file transfer.

    Args:
        total_size: int
        action: str that will be prepended to the progressbar.
            i.e "Uploading: " or "Downloading: "

    Returns:
        progressbar.ProgressBar instance
    """
    return progressbar.ProgressBar(
        max_value=total_size,
        widgets=[
            action,
            progressbar.Percentage(),
            " ",
            progressbar.Bar(marker="#", left="[", right="]"),
            " ",
            progressbar.ETA(),
            " ",
            progressbar.Timer(),
            " ",
            progressbar.FileTransferSpeed(),
        ],
        redirect_stdout=True,
    )


def get_regular_progress_bar(total_size, action):
    """Get progressbar.ProgressBar instance.

    Args:
        total_size: int
        action: str that will be prepended to the progressbar.
            i.e "Uploading: " or "Downloading: "

    Returns:
        progressbar.ProgressBar instance
    """
    return progressbar.ProgressBar(
        max_value=total_size,
        redirect_stdout=True,
        widgets=[
            action,
            progressbar.Percentage(),
            " ",
            progressbar.Bar(marker="#", left="[", right="]"),
            " ",
            progressbar.ETA(),
            " ",
            progressbar.Timer(),
        ],
    )


def validate_credentials(credentials):
    """Validate user credentials."""
    if credentials.email and credentials.password and credentials.api_key:
        echo_debug("User provided 2 sets of credentials.")
        echo_warning(
            "Multiple sets of credentials provided."
            "Please provide either username/password or API key."
        )
        return False

    return True


def login(api_client, credentials):
    """Login user into Gencove's system."""
    if credentials.api_key:
        echo_debug("User authorized via api key")
        api_client.set_api_key(credentials.api_key)
        return True

    if not credentials.email or not credentials.password:
        echo_info("Login required")
        if not credentials.email:
            credentials.email = click.prompt("Email", type=str, err=True)
        if not credentials.password:
            credentials.password = click.prompt(
                "Password", type=str, hide_input=True, err=True
            )
    try:
        api_client.login(credentials.email, credentials.password, credentials.otp_token)
        echo_debug("User logged in successfully")
        return True
    except APIClientError as err:
        if "otp_token" in err.message:
            echo_info("One time password required")
            credentials.otp_token = click.prompt(
                "One time password", type=str, err=True
            )
            return login(api_client, credentials)
        echo_debug(f"Failed to login: {err}")
        echo_error("Failed to login. Please verify your credentials and try again")
        return False


def batchify(items_list, batch_size=500):
    """Generate batches from items list.

    Args:
        items_list (list): list that will be batchified.
        batch_size (int, default=500): batch size that will be returned.
            last batch is not promised to be exactly the length of batch_size.

    Returns:
        subset of items_list
    """
    total = len(items_list)
    left_to_process = total
    start = 0
    while left_to_process >= 0:
        end = start + batch_size
        end = min(end, total)

        yield items_list[start:end]
        start += batch_size
        left_to_process -= batch_size


def enum_as_dict(enum):
    """Convert enum to dict.

    Args:
        enum (Enum): Enumeration to be converted to dict.

    Returns:
        dict Dictionary representation of enum.
    """
    return {s.name: s.value for s in enum}


def python_version_check():
    """Log a warning if the user is using a deprecated version of Python"""
    major_version, minor_version, micro_version = (
        sys.version_info.major,
        sys.version_info.minor,
        sys.version_info.micro,
    )

    if major_version < MINIMUM_SUPPORTED_PYTHON_MAJOR or (
        major_version == MINIMUM_SUPPORTED_PYTHON_MAJOR
        and minor_version < MINIMUM_SUPPORTED_PYTHON_MINOR
    ):
        echo_warning(
            f"Your Python version ({major_version}.{minor_version}.{micro_version}) "
            f"is out of date. Note that the Gencove CLI will drop support for Python "
            f"versions below "
            f"{MINIMUM_SUPPORTED_PYTHON_MAJOR}.{MINIMUM_SUPPORTED_PYTHON_MINOR} "
            f"in the near future. Please upgrade your Python distribution at "
            f"your convenience."
        )



"""Test uploads list command."""
# pylint: disable=wrong-import-order, import-error

from click.testing import CliRunner

from gencove.client import APIClient, APIClientError, APIClientTimeout
from gencove.command.uploads.list.cli import list_uploads
from gencove.models import SampleSheet
from gencove.tests.decorators import assert_authorization
from gencove.tests.filters import filter_jwt, replace_gencove_url_vcr
from gencove.tests.upload.vcr.filters import (
    filter_sample_sheet_response,
    filter_volatile_dates,
)
from gencove.tests.utils import get_vcr_response

import pytest

from vcr import VCR


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/uploads/vcr",
        "filter_headers": [
            "Authorization",
            "Content-Length",
            "User-Agent",
        ],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
        ],
        "before_record_response": [
            filter_jwt,
            filter_volatile_dates,
            filter_sample_sheet_response,
        ],
    }


@pytest.mark.vcr
@assert_authorization
def test_list_does_not_exist(mocker, credentials):
    """Test user cannot get to uploads."""
    runner = CliRunner()
    mocked_get_projects = mocker.patch.object(
        APIClient,
        "get_sample_sheet",
        side_effect=APIClientError(
            message="API Client Error: Not Found: Not found.", status_code=404
        ),
        return_value={"detail": "Not found"},
    )
    res = runner.invoke(list_uploads, credentials)
    assert isinstance(res.exception, SystemExit)
    assert res.exit_code == 1
    mocked_get_projects.assert_called_once()
    assert (
        "\n".join(
            [
                "ERROR: Uploads do not exist.",
                "ERROR: API Client Error: Not Found: Not found.",
                "Aborted!\n",
            ]
        )
        == res.output
    )


@pytest.mark.vcr
@assert_authorization
def test_list_empty(mocker, credentials):
    """Test user has no uploads."""
    runner = CliRunner()
    mocked_get_projects = mocker.patch.object(
        APIClient,
        "get_sample_sheet",
        return_value=SampleSheet(results=[], meta=dict(next=None)),
    )
    res = runner.invoke(list_uploads, credentials)
    assert res.exit_code == 0
    mocked_get_projects.assert_called_once()
    assert "" in res.output


@pytest.mark.vcr
@assert_authorization
def test_list_uploads_slow_response_retry(mocker, credentials):
    """Test list uploads slow response retry."""
    runner = CliRunner()
    mocked_get_sample_sheet = mocker.patch.object(
        APIClient,
        "get_sample_sheet",
        side_effect=APIClientTimeout("Could not connect to the api server"),
    )
    res = runner.invoke(list_uploads, credentials)
    assert res.exit_code == 1
    assert mocked_get_sample_sheet.call_count == 5
    assert res.output == "ERROR: Could not connect to the api server\nAborted!\n"


@pytest.mark.vcr
@assert_authorization
def test_list_uploads(mocker, credentials, recording, vcr):
    """Test list uploads being outputed to the shell."""
    runner = CliRunner()
    if not recording:
        sample_sheet_response = get_vcr_response("/api/v2/sample-sheet/", vcr)
        mocked_get_sample_sheet = mocker.patch.object(
            APIClient,
            "get_sample_sheet",
            return_value=SampleSheet(**sample_sheet_response),
        )
    res = runner.invoke(list_uploads, credentials)
    assert res.exit_code == 0

    if not recording:
        mocked_get_sample_sheet.assert_called()
        uploads = sample_sheet_response["results"]
        uploads = "\n".join(
            [
                "\t".join(
                    [
                        upload["client_id"],
                        upload["fastq"]["r1"]["upload"],
                        upload["fastq"]["r1"]["destination_path"],
                    ]
                )
                for upload in uploads
            ]
        )
        assert f"{uploads}\n" == res.output



"""Commands to be executed from command line."""
# pylint: disable=E0012,C0330,R0913
import click

from .autoimports.cli import autoimports
from .s3_import.cli import s3_import


@click.group()
def s3():  # pylint: disable=C0103
    """S3 imports managements commands.
    Setup guide: https://docs.gencove.com/main/s3-imports/
    """


s3.add_command(autoimports)
s3.add_command(s3_import)



"""Create project's batch executor."""
from pathlib import Path

from gencove import client  # noqa: I100
from gencove.command.base import Command
from gencove.command.download.utils import download_file
from gencove.models import SampleManifests


class GetSampleManifests(Command):
    """Get project sample manifests executor."""

    def __init__(
        self,
        project_id,
        destination,
        credentials,
        options,
    ):
        super().__init__(credentials, options)
        self.project_id = project_id
        self.destination = destination

    def initialize(self):
        """Initialize list subcommand."""
        self.login()

    def validate(self):
        """Validate command input.

        Raises:
            ValidationError - if something is wrong with command parameters.
        """

    def execute(self):
        """Request to download all sample manifests from a project."""
        self.echo_debug(f"Retrieving sample manifests for project {self.project_id}")

        try:
            sample_manifests = self.api_client.get_sample_manifests(
                project_id=self.project_id
            )
            sample_manifests_validated = SampleManifests(results=sample_manifests)
            self.echo_debug(sample_manifests_validated)
            for sample_manifest in sample_manifests_validated.results:
                dst = (
                    Path(self.destination)
                    / str(sample_manifest.id)
                    / sample_manifest.file_name
                )
                dst.parent.mkdir(exist_ok=True, parents=True)
                self.echo_info(
                    f"Downloading sample manifest with ID"
                    f" {sample_manifest.id} to {dst}"
                )
                download_file(
                    file_path=str(dst),
                    download_url=sample_manifest.file.download_url,
                    no_progress=True,
                )
        except client.APIClientError as err:
            self.echo_debug(err)
            if err.status_code == 400:
                self.echo_warning(
                    "There was an error retrieving the sample "
                    "manifests for this project."
                )
                self.echo_info("The following error was returned:")
                self.echo_info(err.message)
            elif err.status_code == 404:
                self.echo_warning(f"Project {self.project_id} does not exist.")
            else:
                raise



"""Test delete projects."""
# pylint: disable=wrong-import-order, import-error
import operator
from uuid import uuid4

from click.testing import CliRunner

from gencove.client import (
    APIClient,
    APIClientError,
)  # noqa: I100
from gencove.command.projects.cli import delete_projects
from gencove.exceptions import MaintenanceError
from gencove.tests.decorators import assert_authorization, assert_no_requests
from gencove.tests.filters import filter_jwt, replace_gencove_url_vcr
from gencove.tests.projects.vcr.filters import (
    filter_projects_delete,
)
from gencove.tests.upload.vcr.filters import filter_volatile_dates
from gencove.tests.utils import get_vcr_response

import pytest

from vcr import VCR


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/projects/vcr",
        "filter_headers": [
            "Authorization",
            "Content-Length",
            "User-Agent",
        ],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
            filter_projects_delete,
        ],
        "before_record_response": [
            filter_jwt,
            filter_volatile_dates,
        ],
    }


@assert_no_requests
def test_delete_projects__bad_project_id(
    mocker, credentials
):  # pylint: disable=unused-argument
    """Test delete projects when non-uuid string is used as project id."""
    runner = CliRunner()

    res = runner.invoke(
        delete_projects,
        [
            "11111111-1111-1111-1111-111111111111,22",
            *credentials,
        ],
    )
    assert res.exit_code == 1
    assert "Not all project_ids are valid." in res.output


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_delete_projects__not_owned_project(credentials, mocker):
    """Test delete projects failure when project is not owned."""
    runner = CliRunner()
    mocked_delete_projects = mocker.patch.object(
        APIClient,
        "delete_projects",
        side_effect=APIClientError(message="", status_code=403),
    )

    res = runner.invoke(
        delete_projects,
        [
            str(uuid4()),
            *credentials,
        ],
    )
    assert res.exit_code == 1
    mocked_delete_projects.assert_called_once()
    assert "You do not have the sufficient permission" in res.output


@assert_no_requests
def test_delete_projects__fail__empty_project_ids(
    credentials,
    mocker,  # pylint: disable=unused-argument
):
    """Test delete projects fail when an empty list of project ids is sent."""

    runner = CliRunner()

    res = runner.invoke(
        delete_projects,
        [
            "",
            *credentials,
        ],
    )
    assert res.exit_code == 1
    assert "No project ids provided" in res.output


@pytest.mark.vcr
@assert_authorization
@pytest.mark.parametrize("remove_hyphens", [True, False])
def test_delete_projects__success(  # pylint: disable=too-many-arguments
    credentials, mocker, project_id_delete, recording, vcr, remove_hyphens
):
    """Test delete projects success."""
    runner = CliRunner()
    if not recording:
        # Mock delete_projects only if using the cassettes, since we
        # mock the return value.
        delete_projects_response = get_vcr_response(
            "/api/v2/projects-delete/", vcr, operator.contains
        )
        mocked_delete_projects = mocker.patch.object(
            APIClient,
            "delete_projects",
            return_value=delete_projects_response,
        )

    if remove_hyphens:
        project_id_delete = project_id_delete.replace("-", "")

    res = runner.invoke(
        delete_projects,
        [
            project_id_delete,
            *credentials,
        ],
    )
    assert res.exit_code == 0
    if not recording:
        mocked_delete_projects.assert_called_once()
    assert "The following projects have been deleted successfully" in res.output


@pytest.mark.vcr
@assert_authorization
def test_delete_projects__returns_maintenance_error_503(
    # pylint: disable=too-many-arguments
    credentials,
    mocker,
    project_id_delete,
    recording,
    using_api_key,
    vcr,
):
    """Test delete projects maintenance error."""
    runner = CliRunner()
    if not recording:
        # Mock delete_projects only if using the cassettes, since we
        # mock the return value.

        delete_projects_resp = get_vcr_response(
            "/api/v2/projects-delete/", vcr, operator.contains
        )
        mocked_delete_projects = mocker.patch.object(
            APIClient,
            "delete_projects",
            side_effect=MaintenanceError(
                message=delete_projects_resp["maintenance_message"],
                eta=delete_projects_resp["maintenance_eta"],
            ),
        )

    res = runner.invoke(
        delete_projects,
        [
            project_id_delete,
            *credentials,
        ],
    )
    assert res.exit_code == 1
    if not recording:
        if using_api_key:
            mocked_delete_projects.assert_called_once()
        else:
            mocked_delete_projects.assert_not_called()

    assert (
        "ERROR: Gencove is currently undergoing maintenance and "
        "will return at the given ETA. "
        "Thank you for your patience."
    ) in res.output



"""List file types"""

import backoff

# pylint: disable=wrong-import-order
from gencove import client
from gencove.command.base import Command
from gencove.command.utils import is_valid_uuid
from gencove.exceptions import ValidationError

from .utils import get_line


class ListFileTypes(Command):
    """List file types command executor."""

    # pylint: disable=too-many-arguments
    def __init__(self, project_id, credentials, options):
        super().__init__(credentials, options)
        self.project_id = project_id

    def initialize(self):
        """Initialize list command."""
        self.login()

    def validate(self):
        """Validate command input.

        Raises:
            ValidationError - if something is wrong with command parameters.
        """
        if self.project_id:
            if is_valid_uuid(self.project_id) is False:
                raise ValidationError("Project ID is not valid. Exiting.")

    def execute(self):
        self.echo_debug("Retrieving file types:")

        try:
            for file_types in self.get_file_types():
                sorted_file_types = sorted(file_types, key=lambda ft: ft.key)
                for file_type in sorted_file_types:
                    self.echo_data(get_line(file_type))

        except client.APIClientError as err:
            if err.status_code == 404:
                self.echo_error(f"Project {self.project_id} does not exist.")
            raise

    @backoff.on_exception(
        backoff.expo,
        (client.APIClientTimeout),
        max_tries=2,
        max_time=30,
    )
    def get_file_types(self):
        """Get file types list

        Yields:
            list of file types
        """
        if self.project_id:
            yield self.api_client.get_file_types(project_id=self.project_id).results
        else:
            yield self.api_client.get_file_types().results



"""Import Biosamples from BaseSpace projects to a project subcommand."""

import json

from ....base import Command
from ....utils import is_valid_json
from ..... import client
from .....exceptions import ValidationError


class BaseSpaceImport(Command):
    """BaseSpace import command executor."""

    def __init__(self, basespace_project_ids, project_id, credentials, options):
        super().__init__(credentials, options)
        self.basespace_project_ids = basespace_project_ids
        self.project_id = project_id
        self.metadata_json = options.metadata_json

    def initialize(self):
        """Initialize basespace-import subcommand."""
        self.login()

    def validate(self):
        """Validate command input.

        Raises:
            ValidationError - if something is wrong with command parameters.
        """
        if self.metadata_json and is_valid_json(self.metadata_json) is False:
            raise ValidationError("Metadata JSON is not valid. Exiting.")

    # no retry for timeouts in order to avoid duplicate heavy operations on
    # the backend
    def execute(self):
        """Make a request to import Biosamples from BaseSpace to a given
        project.
        """
        self.echo_debug(
            "Import Biosamples of the projects {self.basespace_project_ids} "
            f"to a project {self.project_id}"
        )

        try:
            # prepare metadata
            metadata = None
            if self.metadata_json is not None:
                metadata = json.loads(self.metadata_json)
                self.echo_info("Assigning metadata to the imported Biosamples.")
            import_basespace_projects = self.api_client.import_basespace_projects(
                basespace_project_ids=self.basespace_project_ids,
                project_id=self.project_id,
                metadata=metadata,
            )
            self.echo_debug(import_basespace_projects)
            self.echo_info("Request to import BaseSpace projects accepted.")
        except client.APIClientError:
            self.echo_error("There was an error importing Biosamples from BaseSpace.")
            raise



"""Download project's merged VCF file executor."""
import backoff

from ... import download
from ...base import Command
from .... import client
from ....exceptions import ValidationError


class GetMergedVCF(Command):
    """Download project's merged VCF file executor."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        project_id,
        output_filename,
        credentials,
        options,
        no_progress,
    ):
        super().__init__(credentials, options)
        self.project_id = project_id
        self.output_filename = output_filename
        self.no_progress = no_progress

    def initialize(self):
        """Initialize subcommand."""
        self.login()

    def validate(self):
        """Validate command input."""

    @backoff.on_exception(
        backoff.expo,
        (client.APIClientTimeout),
        max_tries=5,
        max_time=30,
    )
    def execute(self):
        """Download merged VCF file for a given project."""
        self.echo_debug(f"Downloading merged VCF file for project {self.project_id}")
        try:
            project = self.api_client.get_project(self.project_id)
            self.echo_debug(project)
            merged_vcf = next(
                (f for f in project.files if f.file_type == "impute-vcf-merged"),
                None,
            )
            if merged_vcf is None:
                raise ValidationError(
                    f"No files to process for project {self.project_id}"
                )
            download_path = (
                self.output_filename
                if self.output_filename
                else download.utils.get_filename_from_download_url(
                    merged_vcf.download_url
                )
            )
            download.utils.download_file(
                download_path,
                merged_vcf.download_url,
                no_progress=self.no_progress,
            )
        except client.APIClientError as err:
            self.echo_debug(err)
            if err.status_code == 404:
                self.echo_error(f"Project {self.project_id} does not exist.")
            raise



"""Utilities for processing and validating samples."""


from gencove.constants import SampleArchiveStatus


def get_line(sample, include_run=False):
    """Build a list of relevant data to be printed.

    Args:
        sample (dict): an object from project samples

    Returns:
        list(str): list of relevant data to be printed
    """
    columns = [
        sample.last_status.created.isoformat(),
        str(sample.id),
        str(sample.client_id),
        sample.last_status.status,
        sample.archive_last_status.status
        if sample.archive_last_status is not None
        else SampleArchiveStatus.UNKNOWN.value,
    ]
    if include_run:
        columns.append(str(sample.run))

    return "\t".join(columns)



"""List Biosamples from BaseSpace project subcommand."""
import backoff

from .utils import get_line
from ....base import Command
from ..... import client


class BiosamplesList(Command):
    """Biosamples list command executor."""

    def __init__(self, basespace_project_id, credentials, options):
        super().__init__(credentials, options)
        self.basespace_project_id = basespace_project_id

    def initialize(self):
        """Initialize basespace biosamples list subcommand."""
        self.login()

    def validate(self):
        """Need to override because it's not implemented in the base class."""
        pass  # pylint: disable=unnecessary-pass

    def execute(self):
        """Make a request to list Biosamples from a BaseSpace project."""
        self.echo_debug(
            f"Listing Biosamples from a BaseSpace project {self.basespace_project_id}."
        )
        try:
            for biosamples in self.get_paginated_biosamples():
                if not biosamples:
                    self.echo_debug("No BaseSpace Biosamples were found.")
                    return

                for biosample in biosamples:
                    self.echo_data(get_line(biosample))
        except client.APIClientError as err:
            self.echo_error("There was an error listing Biosamples.")
            if err.status_code == 404:
                self.echo_error("No BaseSpace Biosamples found.")
            raise

    def get_paginated_biosamples(self):
        """Paginate over all Biosamples returned.

        Yields:
            paginated lists of BaseSpace Biosamples
        """
        more = True
        next_link = None
        while more:
            self.echo_debug("Get BaseSpace Biosamples page")
            resp = self.get_biosamples(next_link)

            yield resp.results
            next_link = resp.meta.next
            more = next_link is not None

    @backoff.on_exception(
        backoff.expo,
        (client.APIClientTimeout),
        max_tries=2,
        max_time=30,
    )
    def get_biosamples(self, next_link=None):
        """Get BaseSpace Biosamples page."""
        return self.api_client.list_biosamples(self.basespace_project_id, next_link)



"""Project autoimport from S3 URI shell command definition.
"""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials, Optionals

from .main import S3AutoImportList


@click.command("list")
@add_options(common_options)
def autoimport_list(
    host,
    email,
    password,
    api_key,
):
    """Lists S3 automatic import jobs.

    Examples:

        List S3 automatic import jobs:

            gencove s3 autoimports list
    """  # noqa: E501

    S3AutoImportList(
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
    ).run()



"""Request project delete samples"""
from gencove import client  # noqa: I100
from gencove.command.base import Command


class DeleteSamples(Command):
    """Delete project's samples."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        project_id,
        sample_ids,
        credentials,
        options,
    ):
        super().__init__(credentials, options)
        self.project_id = project_id
        self.sample_ids = sample_ids

    def initialize(self):
        """Initialize list subcommand."""
        self.login()

    def validate(self):
        """Validate command input.

        Raises:
            ValidationError - if something is wrong with command parameters.
        """

    def execute(self):
        """Make a request to request samples delete for given project."""
        self.echo_debug(
            f"Requesting to delete samples in project {self.project_id} for "
            f"samples {self.sample_ids}"
        )

        try:
            deleted_project_samples_details = self.api_client.delete_project_samples(
                project_id=self.project_id,
                sample_ids=self.sample_ids,
            )

            self.echo_debug(deleted_project_samples_details)
            self.echo_info("The following samples have been deleted successfully:")
            for sample in self.sample_ids:
                self.echo_info(f"\t{sample}")
        except client.APIClientError as err:
            self.echo_debug(err)
            if err.status_code == 400:
                self.echo_warning(
                    "There was an error requesting delete project samples."
                )
                self.echo_info("The following error was returned:")
                self.echo_info(err.message)
            elif err.status_code == 404:
                self.echo_warning(f"Project {self.project_id} does not exist.")
            else:
                raise



"""Test download command."""
# pylint: disable=wrong-import-order, import-error, too-many-lines
import io
import json
import operator
import os
import sys
from unittest.mock import call
from uuid import UUID, uuid4

from click import echo
from click.testing import CliRunner

from gencove.cli import download
from gencove.client import APIClient
from gencove.command.download.utils import download_file
from gencove.models import (
    ProjectSamples,
    SampleDetails,
    SampleMetadata,
    SampleQC,
)
from gencove.tests.decorators import assert_authorization
from gencove.tests.download.vcr.filters import (
    filter_files_request,
    filter_files_response,
    filter_project_samples_request,
)
from gencove.tests.filters import (
    filter_aws_headers,
    filter_file_types_request,
    filter_jwt,
    filter_project_samples_response,
    filter_sample_metadata_request,
    filter_sample_quality_controls_request,
    filter_samples_request,
    filter_samples_response,
    mock_binary_response,
    replace_gencove_url_vcr,
    replace_s3_from_url,
)
from gencove.tests.utils import MOCK_CHECKSUM, MOCK_UUID, get_vcr_response

from pydantic.networks import HttpUrl

import pytest

from vcr import VCR


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/download/vcr",
        "filter_headers": [
            "Authorization",
            "Content-Length",
            "User-Agent",
        ],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
            filter_file_types_request,
            filter_files_request,
            filter_project_samples_request,
            filter_samples_request,
            filter_sample_quality_controls_request,
            filter_sample_metadata_request,
            replace_s3_from_url,
        ],
        "before_record_response": [
            filter_files_response,
            filter_jwt,
            filter_project_samples_response,
            filter_samples_response,
            mock_binary_response,
            filter_aws_headers,
        ],
    }


@pytest.mark.vcr
@assert_authorization
def test_no_required_options(credentials, mocker):  # pylint: disable=unused-argument
    """Test that command exits without project id or sample id provided."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        res = runner.invoke(
            download,
            [
                "cli_test_data",
                "--project-id",
                "11111111-1111-1111-1111-111111111111",
                *credentials,
            ],
        )
        assert res.exit_code == 1


@pytest.mark.vcr
def test_both_project_id_and_sample_ids_provided(credentials, mocker):
    """Command exits if both project id and sample ids are provided."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        mocked_login = mocker.patch.object(APIClient, "login")
        res = runner.invoke(
            download,
            [
                "cli_test_data",
                "--project-id",
                "11111111-1111-1111-1111-111111111111",
                "--sample-ids",
                "1,2,3",
                *credentials,
            ],
        )
    assert res.exit_code == 1
    mocked_login.assert_not_called()


@pytest.mark.vcr
@assert_authorization
def test_project_id_provided(credentials, mocker, project_id_download, recording, vcr):
    """Check happy flow."""
    # pylint: disable=too-many-locals
    runner = CliRunner()
    with runner.isolated_filesystem():
        if not recording:
            # Mock only if using the cassettes, since we mock the return value.
            get_project_samples_response = get_vcr_response(
                "/api/v2/project-samples/", vcr, operator.contains
            )
            mocked_project_samples = mocker.patch.object(
                APIClient,
                "get_project_samples",
                return_value=ProjectSamples(**get_project_samples_response),
            )
            get_sample_details_response = get_vcr_response(
                "/api/v2/samples/", vcr, operator.contains
            )
            mocked_sample_details = mocker.patch.object(
                APIClient,
                "get_sample_details",
                return_value=SampleDetails(**get_sample_details_response),
            )
            get_sample_qc_metrics_response = get_vcr_response(
                "/api/v2/sample-quality-controls/", vcr, operator.contains
            )
            mocked_qc_metrics = mocker.patch.object(
                APIClient,
                "get_sample_qc_metrics",
                return_value=SampleQC(**get_sample_qc_metrics_response),
            )
            get_metadata_response = get_vcr_response(
                "/api/v2/sample-metadata/", vcr, operator.contains
            )
            mocked_get_metadata = mocker.patch.object(
                APIClient,
                "get_metadata",
                return_value=SampleMetadata(**get_metadata_response),
            )
        mocked_download_file = mocker.patch(
            "gencove.command.download.main.download_file",
            side_effect=download_file,
        )
        res = runner.invoke(
            download,
            [
                "cli_test_data",
                "--project-id",
                project_id_download,
                *credentials,
            ],
        )
        assert res.exit_code == 0
        if not recording:
            mocked_project_samples.assert_called_once()
            mocked_sample_details.assert_called_once()
            mocked_qc_metrics.assert_called_once()
            mocked_get_metadata.assert_called_once()
        assert mocked_download_file.call_count == 2


@pytest.mark.vcr
@assert_authorization
def test_sample_ids_provided(credentials, mocker, recording, sample_id_download, vcr):
    """Check happy flow with sample ids."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        if not recording:
            # Mock only if using the cassettes, since we mock the return value.
            get_sample_details_response = get_vcr_response(
                "/api/v2/samples/", vcr, operator.contains
            )
            mocked_sample_details = mocker.patch.object(
                APIClient,
                "get_sample_details",
                return_value=SampleDetails(**get_sample_details_response),
            )
            get_sample_qc_metrics_response = get_vcr_response(
                "/api/v2/sample-quality-controls/", vcr, operator.contains
            )
            mocked_qc_metrics = mocker.patch.object(
                APIClient,
                "get_sample_qc_metrics",
                return_value=SampleQC(**get_sample_qc_metrics_response),
            )
            get_metadata_response = get_vcr_response(
                "/api/v2/sample-metadata/", vcr, operator.contains
            )
            mocked_get_metadata = mocker.patch.object(
                APIClient,
                "get_metadata",
                return_value=SampleMetadata(**get_metadata_response),
            )
            mocked_download_file = mocker.patch(
                "gencove.command.download.main.download_file",
                side_effect=download_file,
            )
        res = runner.invoke(
            download,
            [
                "cli_test_data",
                "--sample-ids",
                sample_id_download,
                *credentials,
            ],
        )
        assert res.exit_code == 0
        if not recording:
            mocked_sample_details.assert_called_once()
            mocked_qc_metrics.assert_called_once()
            mocked_get_metadata.assert_called_once()
            calls = [
                call(
                    f"cli_test_data/mock_client_id/{MOCK_UUID}/r1.fastq.gz",
                    HttpUrl(
                        url="https://example.com/r1.fastq.gz",
                        scheme="https",
                        host="example.com",
                    ),
                    True,
                    False,
                ),
                call(
                    f"cli_test_data/mock_client_id/{MOCK_UUID}/r2.fastq.gz",
                    HttpUrl(
                        url="https://example.com/r2.fastq.gz",
                        scheme="https",
                        host="example.com",
                    ),
                    True,
                    False,
                ),
            ]
            mocked_download_file.assert_has_calls(calls)


@pytest.mark.vcr
@assert_authorization
def test_sample_ids_provided_no_qc_file(
    credentials, mocker, recording, sample_id_download, vcr
):
    """Check flow with sample ids and no file present."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        mocked_sample_details = mocker.patch.object(
            APIClient,
            "get_sample_details",
            return_value=SampleDetails(
                **{
                    "id": sample_id_download,
                    "client_id": "1",
                    "last_status": {
                        "id": str(uuid4()),
                        "status": "succeeded",
                        "created": "2020-07-28T12:46:22.719862Z",
                    },
                    "archive_last_status": {
                        "id": str(uuid4()),
                        "status": "available",
                        "created": "2020-07-28T12:46:22.719862Z",
                        "transition_cutoff": "2020-08-28T12:46:22.719862Z",
                    },
                    "files": [],
                }
            ),
        )
        if not recording:
            # Mock only if using the cassettes, since we mock the return value.
            get_sample_qc_metrics_response = get_vcr_response(
                "/api/v2/sample-quality-controls/", vcr, operator.contains
            )
            mocked_qc_metrics = mocker.patch.object(
                APIClient,
                "get_sample_qc_metrics",
                return_value=SampleQC(**get_sample_qc_metrics_response),
            )
            get_metadata_response = get_vcr_response(
                "/api/v2/sample-metadata/", vcr, operator.contains
            )
            mocked_get_metadata = mocker.patch.object(
                APIClient,
                "get_metadata",
                return_value=SampleMetadata(**get_metadata_response),
            )
        mocked_download_file = mocker.patch(
            "gencove.command.download.main.download_file",
            side_effect=download_file,
        )
        res = runner.invoke(
            download,
            [
                "cli_test_data",
                "--sample-ids",
                sample_id_download,
                *credentials,
            ],
        )
        assert res.exit_code == 0
        mocked_download_file.assert_not_called()
        if not recording:
            mocked_sample_details.assert_called_once()
            mocked_qc_metrics.assert_called_once()
            mocked_get_metadata.assert_called_once()


@pytest.mark.vcr
@assert_authorization
def test_create_checksum_file(credentials, mocker, recording, sample_id_download, vcr):
    """Check checksums flag."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        if not recording:
            # Mock only if using the cassettes, since we mock the return value.
            get_sample_details_response = get_vcr_response(
                "/api/v2/samples/", vcr, operator.contains
            )
            mocked_sample_details = mocker.patch.object(
                APIClient,
                "get_sample_details",
                return_value=SampleDetails(**get_sample_details_response),
            )
            mocked_download_file = mocker.patch(
                "gencove.command.download.main.download_file",
                side_effect=download_file,
            )
            get_file_checksum_response = get_vcr_response(
                "/api/v2/files/", vcr, operator.contains
            )
            mocked_get_file_checksum = mocker.patch.object(
                APIClient,
                "get_file_checksum",
                return_value=get_file_checksum_response.decode(),
            )
        res = runner.invoke(
            download,
            [
                "cli_test_data",
                "--sample-ids",
                sample_id_download,
                *credentials,
                "--file-types",
                "fastq-r2",
                "--checksums",
            ],
        )
        assert res.exit_code == 0
        if not recording:
            mocked_sample_details.assert_called_once()
            mocked_download_file.assert_called_once_with(
                f"cli_test_data/mock_client_id/{MOCK_UUID}/r2.fastq.gz",
                HttpUrl(
                    url="https://example.com/r2.fastq.gz",
                    scheme="https",
                    host="example.com",
                ),
                True,
                False,
            )
            mocked_get_file_checksum.assert_called_once_with(UUID(MOCK_UUID))
            checksum_path = (
                f"cli_test_data/mock_client_id/{MOCK_UUID}/r2.fastq.gz.sha256"
            )
            assert os.path.exists(checksum_path)
            with open(checksum_path, "r", encoding="utf-8") as checksum_file:
                assert (
                    checksum_file.read()
                    == f"{MOCK_CHECKSUM} *{MOCK_UUID}_R2.fastq.gz\n"
                )


@pytest.mark.vcr
@assert_authorization
def test_create_checksum_file_exception(
    credentials, mocker, recording, sample_id_download, vcr
):
    """Check that checksums flag fails if file doesn't have checksum
    (r1 file).
    """
    runner = CliRunner()
    with runner.isolated_filesystem():
        if not recording:
            # Mock only if using the cassettes, since we mock the return value.
            get_sample_details_response = get_vcr_response(
                "/api/v2/samples/", vcr, operator.contains
            )
            mocked_sample_details = mocker.patch.object(
                APIClient,
                "get_sample_details",
                return_value=SampleDetails(**get_sample_details_response),
            )
            mocked_download_file = mocker.patch(
                "gencove.command.download.main.download_file",
                side_effect=download_file,
            )
        res = runner.invoke(
            download,
            [
                "cli_test_data",
                "--sample-ids",
                sample_id_download,
                *credentials,
                "--file-types",
                "fastq-r1",
                "--checksums",
            ],
        )
        assert res.exit_code == 1
        assert (
            "ERROR: API Client Error: Bad Request: File does not have checksum."  # noqa: E501 pylint: disable=line-too-long
            in res.stdout
        )
        if not recording:
            mocked_sample_details.assert_called_once()
            mocked_download_file.assert_called_once_with(
                f"cli_test_data/mock_client_id/{MOCK_UUID}/r1.fastq.gz",
                HttpUrl(
                    url="https://example.com/r1.fastq.gz",
                    scheme="https",
                    host="example.com",
                ),
                True,
                False,
            )
            file_path = f"cli_test_data/mock_client_id/{MOCK_UUID}/r1.fastq.gz"
            checksum_path = f"{file_path}.sha256"
            assert not os.path.exists(checksum_path)


def test_multiple_credentials_not_allowed(mocker):
    """Test that in providing multiple credentials is not allowed."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login")
    res = runner.invoke(
        download,
        [
            "some_test_data",
            "--sample-ids",
            "0",
            "--email",
            "foo@bar.com",
            "--password",
            "123",
            "--api-key",
            "foobar",
        ],
    )
    assert res.exit_code == 1
    assert "Please provide either username/password or API key." in res.output
    mocked_login.assert_not_called()


@pytest.mark.vcr
@assert_authorization
def test_download_stdout_no_flag(
    credentials, mocker, project_id_download, recording, vcr
):
    """Test command exits if no flag provided and stdout defined."""
    runner = CliRunner()
    if not recording:
        # Mock only if using the cassettes, since we mock the return value.
        get_project_samples_response = get_vcr_response(
            "/api/v2/project-samples/", vcr, operator.contains
        )
        mocked_project_samples = mocker.patch.object(
            APIClient,
            "get_project_samples",
            return_value=ProjectSamples(**get_project_samples_response),
        )
    res = runner.invoke(
        download,
        [
            "-",
            "--project-id",
            project_id_download,
            *credentials,
        ],
    )
    assert res.exit_code == 1
    output_line = io.BytesIO()
    sys.stdout = output_line
    echo("ERROR: Cannot have - as a destination without download-urls.")
    assert output_line.getvalue() in res.output.encode()
    if not recording:
        mocked_project_samples.assert_called_once()


@pytest.mark.vcr
@assert_authorization
def test_download_stdout_with_flag(
    credentials, mocker, project_id_download, recording, vcr
):
    """Test command outputs json to stdout."""
    # pylint: disable=too-many-locals
    runner = CliRunner()
    if not recording:
        # Mock only if using the cassettes, since we mock the return value.
        get_project_samples_response = get_vcr_response(
            "/api/v2/project-samples/", vcr, operator.contains
        )
        mocked_project_samples = mocker.patch.object(
            APIClient,
            "get_project_samples",
            return_value=ProjectSamples(**get_project_samples_response),
        )
        get_sample_details_response = get_vcr_response(
            "/api/v2/samples/", vcr, operator.contains
        )
        sample = SampleDetails(**get_sample_details_response)
        mocked_sample_details = mocker.patch.object(
            APIClient,
            "get_sample_details",
            return_value=sample,
        )
    res = runner.invoke(
        download,
        [
            "-",
            "--project-id",
            project_id_download,
            *credentials,
            "--download-urls",
        ],
    )
    assert res.exit_code == 0
    if not recording:
        mocked_project_samples.assert_called_once()
        mocked_sample_details.assert_called_once()
        output_line = io.BytesIO()
        sys.stdout = output_line
        for _ in get_project_samples_response["results"]:
            archive_last_status_created = sample.archive_last_status.created.isoformat()
            download_url_1 = "https://example.com/r1.fastq.gz"
            download_url_2 = "https://example.com/r2.fastq.gz"
            mocked_result = json.dumps(
                [
                    {
                        "gencove_id": MOCK_UUID,
                        "client_id": "mock_client_id",
                        "last_status": {
                            "id": MOCK_UUID,
                            "status": sample.last_status.status,
                            "created": sample.last_status.created.isoformat(),
                        },
                        "archive_last_status": {
                            "id": MOCK_UUID,
                            "status": sample.archive_last_status.status,
                            "created": archive_last_status_created,
                            "transition_cutoff": None,
                        },
                        "files": {
                            "fastq-r1": {
                                "id": MOCK_UUID,
                                "download_url": download_url_1,
                                "checksum_sha256": "",
                            },
                            "fastq-r2": {
                                "id": MOCK_UUID,
                                "download_url": download_url_2,
                                "checksum_sha256": MOCK_CHECKSUM,
                            },
                        },
                    }
                ],
                indent=4,
            )
            echo(mocked_result)
        assert output_line.getvalue() in res.output.encode()


@pytest.mark.vcr
@assert_authorization
def test_download_urls_to_file(
    credentials, mocker, project_id_download, recording, vcr
):
    """Test saving downloaded urls output to a json file."""
    # pylint: disable=too-many-locals
    runner = CliRunner()
    if not recording:
        # Mock only if using the cassettes, since we mock the return value.
        get_project_samples_response = get_vcr_response(
            "/api/v2/project-samples/", vcr, operator.contains
        )
        mocked_project_samples = mocker.patch.object(
            APIClient,
            "get_project_samples",
            return_value=ProjectSamples(**get_project_samples_response),
        )
        get_sample_details_response = get_vcr_response(
            "/api/v2/samples/", vcr, operator.contains
        )
        sample = SampleDetails(**get_sample_details_response)
        mocked_sample_details = mocker.patch.object(
            APIClient,
            "get_sample_details",
            return_value=sample,
        )
    with runner.isolated_filesystem():
        res = runner.invoke(
            download,
            [
                "output.json",
                "--project-id",
                project_id_download,
                *credentials,
                "--download-urls",
            ],
        )
        assert res.exit_code == 0
        output_file = []
        if not recording:
            for _ in get_project_samples_response["results"]:
                archive_last_status_created = (
                    sample.archive_last_status.created.isoformat()
                )
                download_url_1 = "https://example.com/r1.fastq.gz"
                download_url_2 = "https://example.com/r2.fastq.gz"
                output_file.append(
                    {
                        "gencove_id": MOCK_UUID,
                        "client_id": "mock_client_id",
                        "last_status": {
                            "id": MOCK_UUID,
                            "status": sample.last_status.status,
                            "created": sample.last_status.created.isoformat(),
                        },
                        "archive_last_status": {
                            "id": MOCK_UUID,
                            "status": sample.archive_last_status.status,
                            "created": archive_last_status_created,
                            "transition_cutoff": None,
                        },
                        "files": {
                            "fastq-r1": {
                                "id": MOCK_UUID,
                                "download_url": download_url_1,
                                "checksum_sha256": "",
                            },
                            "fastq-r2": {
                                "id": MOCK_UUID,
                                "download_url": download_url_2,
                                "checksum_sha256": MOCK_CHECKSUM,
                            },
                        },
                    }
                )
            with open("output.json", "r", encoding="utf-8") as json_file:
                assert output_file == json.load(json_file)
            mocked_project_samples.assert_called_once()
            mocked_sample_details.assert_called_once()


@pytest.mark.vcr
@assert_authorization
def test_download_no_progress(credentials, mocker, recording, sample_id_download, vcr):
    """Test command doesn't show progress bar."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        if not recording:
            # Mock only if using the cassettes, since we mock the return value.
            get_sample_details_response = get_vcr_response(
                "/api/v2/samples/", vcr, operator.contains
            )
            mocked_sample_details = mocker.patch.object(
                APIClient,
                "get_sample_details",
                return_value=SampleDetails(**get_sample_details_response),
            )
            get_sample_qc_metrics_response = get_vcr_response(
                "/api/v2/sample-quality-controls/", vcr, operator.contains
            )
            mocked_qc_metrics = mocker.patch.object(
                APIClient,
                "get_sample_qc_metrics",
                return_value=SampleQC(**get_sample_qc_metrics_response),
            )
            get_metadata_response = get_vcr_response(
                "/api/v2/sample-metadata/", vcr, operator.contains
            )
            mocked_get_metadata = mocker.patch.object(
                APIClient,
                "get_metadata",
                return_value=SampleMetadata(**get_metadata_response),
            )
        mocked_download_file = mocker.patch(
            "gencove.command.download.main.download_file",
            side_effect=download_file,
        )
        res = runner.invoke(
            download,
            [
                "cli_test_data",
                "--sample-ids",
                sample_id_download,
                *credentials,
                "--no-progress",
            ],
        )
        assert res.exit_code == 0
        if not recording:
            mocked_download_file.assert_has_calls(
                [
                    call(
                        f"cli_test_data/mock_client_id/{MOCK_UUID}/r1.fastq.gz",  # noqa: E501 pylint: disable=line-too-long
                        HttpUrl(
                            url="https://example.com/r1.fastq.gz",
                            scheme="https",
                            host="example.com",
                        ),
                        True,
                        True,
                    ),
                    call(
                        f"cli_test_data/mock_client_id/{MOCK_UUID}/r2.fastq.gz",  # noqa: E501 pylint: disable=line-too-long
                        HttpUrl(
                            url="https://example.com/r2.fastq.gz",
                            scheme="https",
                            host="example.com",
                        ),
                        True,
                        True,
                    ),
                ]
            )
            mocked_sample_details.assert_called_once()
            mocked_qc_metrics.assert_called_once()
            mocked_get_metadata.assert_called_once()


@pytest.mark.vcr
@assert_authorization
def test_project_id_provided_skip_existing_qc_and_metadata(
    credentials, mocker, project_id_download, recording, vcr
):
    """Check happy flow."""
    # pylint: disable=too-many-locals
    runner = CliRunner()
    with runner.isolated_filesystem():
        if not recording:
            # Mock only if using the cassettes, since we mock the return value.
            get_project_samples_response = get_vcr_response(
                "/api/v2/project-samples/", vcr, operator.contains
            )
            mocked_project_samples = mocker.patch.object(
                APIClient,
                "get_project_samples",
                return_value=ProjectSamples(**get_project_samples_response),
            )
            get_sample_details_response = get_vcr_response(
                "/api/v2/samples/", vcr, operator.contains
            )
            mocked_sample_details = mocker.patch.object(
                APIClient,
                "get_sample_details",
                return_value=SampleDetails(**get_sample_details_response),
            )
            get_sample_qc_metrics_response = get_vcr_response(
                "/api/v2/sample-quality-controls/", vcr, operator.contains
            )
            mocked_qc_metrics = mocker.patch.object(
                APIClient,
                "get_sample_qc_metrics",
                return_value=SampleQC(**get_sample_qc_metrics_response),
            )
            get_metadata_response = get_vcr_response(
                "/api/v2/sample-metadata/", vcr, operator.contains
            )
            mocked_get_metadata = mocker.patch.object(
                APIClient,
                "get_metadata",
                return_value=SampleMetadata(**get_metadata_response),
            )
        mocked_download_file = mocker.patch(
            "gencove.command.download.main.download_file",
            side_effect=download_file,
        )
        res = runner.invoke(
            download,
            [
                "cli_test_data",
                "--project-id",
                project_id_download,
                *credentials,
                "--no-skip-existing",
            ],
        )
        assert res.exit_code == 0
        if not recording:
            mocked_project_samples.assert_called_once()
            mocked_sample_details.assert_called_once()
            mocked_qc_metrics.assert_called_once()
            mocked_get_metadata.assert_called_once()
        assert mocked_download_file.call_count == 2

        # call it the second time
        mocked_qc_metrics = mocker.patch.object(APIClient, "get_sample_qc_metrics")
        mocked_get_metadata = mocker.patch.object(APIClient, "get_metadata")
        res = runner.invoke(
            download,
            [
                "cli_test_data",
                "--project-id",
                project_id_download,
                *credentials,
                "--skip-existing",
            ],
        )
        assert res.exit_code == 0
        mocked_qc_metrics.assert_not_called()
        mocked_get_metadata.assert_not_called()


@pytest.mark.vcr
@assert_authorization
def test_download_not_working_because_archived(
    archived_sample, credentials, mocker, recording, vcr
):
    """Test command doesn't download archived files."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        if not recording:
            # Mock only if using the cassettes, since we mock the return value.
            get_sample_details_response = get_vcr_response(
                "/api/v2/samples/", vcr, operator.contains
            )
            mocked_sample_details = mocker.patch.object(
                APIClient,
                "get_sample_details",
                return_value=SampleDetails(**get_sample_details_response),
            )
        mocked_qc_metrics = mocker.patch.object(APIClient, "get_sample_qc_metrics")
        mocked_get_metadata = mocker.patch.object(APIClient, "get_metadata")
        mocked_download_file = mocker.patch(
            "gencove.command.download.main.download_file",
        )
        res = runner.invoke(
            download,
            [
                "cli_test_data",
                "--sample-ids",
                archived_sample,
                *credentials,
            ],
        )
        assert res.exit_code == 1
        assert "is archived and cannot be downloaded " in res.output
        if not recording:
            mocked_sample_details.assert_called_once()
        mocked_qc_metrics.assert_not_called()
        mocked_get_metadata.assert_not_called()
        mocked_download_file.assert_not_called()


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_project_id_provided_filter_not_archived(credentials, mocker):
    """Check happy flow."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        mocked_project_samples = mocker.patch.object(
            APIClient,
            "get_project_samples",
            return_value=ProjectSamples(**{"results": [], "meta": {"next": None}}),
        )
        mocked_sample_details = mocker.patch.object(APIClient, "get_sample_details")
        mocked_qc_metrics = mocker.patch.object(APIClient, "get_sample_qc_metrics")
        mocked_get_metadata = mocker.patch.object(APIClient, "get_metadata")
        mocked_download_file = mocker.patch(
            "gencove.command.download.main.download_file"
        )
        res = runner.invoke(
            download,
            [
                "cli_test_data",
                "--project-id",
                "123",
                *credentials,
            ],
        )
        assert res.exit_code == 1
        mocked_project_samples.assert_called_once()
        mocked_qc_metrics.assert_not_called()
        mocked_get_metadata.assert_not_called()
        mocked_download_file.assert_not_called()
        mocked_sample_details.assert_not_called()


@pytest.mark.vcr
def test_invalid_file_types_project_id_provided(
    credentials, mocker, project_id_download, recording, vcr
):
    """Fails when an invalid file type is provided for a project"""
    # pylint: disable=too-many-locals
    runner = CliRunner()
    with runner.isolated_filesystem():
        if not recording:
            # Mock only if using the cassettes, since we mock the return value.
            get_project_samples_response = get_vcr_response(
                "/api/v2/project-samples/", vcr, operator.contains
            )
            mocked_project_samples = mocker.patch.object(
                APIClient,
                "get_project_samples",
                return_value=ProjectSamples(**get_project_samples_response),
            )
        mocked_sample_details = mocker.patch.object(APIClient, "get_sample_details")
        mocked_qc_metrics = mocker.patch.object(APIClient, "get_sample_qc_metrics")
        mocked_get_metadata = mocker.patch.object(APIClient, "get_metadata")
        mocked_download_file = mocker.patch(
            "gencove.command.download.main.download_file",
            side_effect=download_file,
        )
        res = runner.invoke(
            download,
            [
                "cli_test_data",
                "--project-id",
                project_id_download,
                *credentials,
                "--file-types",
                "call_capture-forced_vcf",
            ],
        )
        assert res.exit_code == 1
        assert "ERROR: Invalid file types: call_capture-forced_vcf." in res.stdout
        if not recording:
            mocked_project_samples.assert_called_once()
        mocked_sample_details.assert_not_called()
        mocked_qc_metrics.assert_not_called()
        mocked_get_metadata.assert_not_called()
        mocked_download_file.assert_not_called()


@pytest.mark.vcr
def test_invalid_file_types_sample_ids_provided(
    credentials, mocker, sample_id_download, recording, vcr
):
    """Fails when an invalid file type is provided for a sample"""
    # pylint: disable=too-many-locals
    runner = CliRunner()
    with runner.isolated_filesystem():
        if not recording:
            # Mock only if using the cassettes, since we mock the return value.
            get_sample_details_response = get_vcr_response(
                "/api/v2/samples/", vcr, operator.contains
            )
            mocked_sample_details = mocker.patch.object(
                APIClient,
                "get_sample_details",
                return_value=SampleDetails(**get_sample_details_response),
            )
        mocked_qc_metrics = mocker.patch.object(APIClient, "get_sample_qc_metrics")
        mocked_get_metadata = mocker.patch.object(APIClient, "get_metadata")
        mocked_download_file = mocker.patch(
            "gencove.command.download.main.download_file",
            side_effect=download_file,
        )
        res = runner.invoke(
            download,
            [
                "cli_test_data",
                "--sample-ids",
                sample_id_download,
                *credentials,
                "--file-types",
                "fastqr1",
            ],
        )
        assert res.exit_code == 1
        assert "ERROR: Invalid file types: fastqr1." in res.stdout
        mocked_sample_details.assert_not_called()
        mocked_qc_metrics.assert_not_called()
        mocked_get_metadata.assert_not_called()
        mocked_download_file.assert_not_called()



"""Utilities for processing pipelines."""


def get_line(pipeline):
    """Build a list of relevant data to be printed.

    Args:
        pipeline (dict): an object from project pipelines response

    Returns:
        str: tab delimited list of relevant data to be printed
    """
    return "\t".join(
        [
            str(pipeline.id),
            pipeline.version,
        ]
    )



"""Test project's get sample manifest command."""

# pylint: disable=wrong-import-order, import-error
import operator
import os
import uuid

from click.testing import CliRunner

from gencove.client import APIClient, APIClientError  # noqa: I100
from gencove.command.projects.cli import get_sample_manifests
from gencove.models import SampleManifest
from gencove.tests.decorators import assert_authorization, assert_no_requests
from gencove.tests.filters import (
    filter_aws_headers,
    filter_jwt,
    replace_gencove_url_vcr,
)
from gencove.tests.projects.vcr.filters import (
    filter_get_project_sample_manifests_response,
    filter_get_sample_manifest_files_response,
    filter_project_sample_manifest_request,
    filter_sample_manifests_request,
)
from gencove.tests.upload.vcr.filters import filter_volatile_dates
from gencove.tests.utils import get_vcr_response

import pytest

from vcr import VCR


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/projects/vcr",
        "filter_headers": ["Authorization", "Content-Length", "User-Agent", "ETag"],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
            filter_project_sample_manifest_request,
            filter_sample_manifests_request,
        ],
        "before_record_response": [
            filter_jwt,
            filter_aws_headers,
            filter_volatile_dates,
            filter_get_project_sample_manifests_response,
            filter_get_sample_manifest_files_response,
        ],
    }


@pytest.mark.vcr
@assert_authorization
def test_get_sample_manifests__success(
    credentials,
    mocker,
    project_id_sample_manifest,
    recording,
    vcr,
):  # pylint: disable=unused-argument
    """Test get sample manifests success case"""
    runner = CliRunner()
    if not recording:
        get_sample_manifest_response = get_vcr_response(
            "/api/v2/project-sample-manifests/", vcr, operator.contains
        )
        mocker.patch(
            "gencove.command.projects.get_sample_manifests.main.download_file",
        )
        mocked_get_sample_manifests = mocker.patch.object(
            APIClient,
            "get_sample_manifests",
            return_value=[SampleManifest(**x) for x in get_sample_manifest_response],
        )

    with runner.isolated_filesystem():
        os.mkdir("test_dir")
        res = runner.invoke(
            get_sample_manifests,
            [
                project_id_sample_manifest,
                "test_dir",
                *credentials,
            ],
        )
    assert res.exit_code == 0
    if not recording:
        mocked_get_sample_manifests.assert_called_once()
    assert "Downloading sample manifest" in res.output


@pytest.mark.vcr
@assert_authorization
def test_get_sample_manifests__empty(
    credentials,
    mocker,
    project_id,  # project without any manifests
    recording,
    vcr,
):
    """Get sample manifests empty case"""
    runner = CliRunner()
    if not recording:
        get_sample_manifest_response = get_vcr_response(
            "/api/v2/project-sample-manifests/", vcr, operator.contains
        )
        mocked_get_sample_manifests = mocker.patch.object(
            APIClient,
            "get_sample_manifests",
            return_value=get_sample_manifest_response,
        )
        mocker.patch(
            "gencove.command.projects.get_sample_manifests.main.download_file",
        )
    with runner.isolated_filesystem():
        os.mkdir("test_dir")
        res = runner.invoke(
            get_sample_manifests,
            [
                project_id,
                "test_dir",
                *credentials,
            ],
        )
    assert res.exit_code == 0
    if not recording:
        mocked_get_sample_manifests.assert_called_once()
    assert res.output == ""


@pytest.mark.vcr
@assert_authorization
def test_get_sample_manifests__not_owned_project(
    credentials, recording, mocker, vcr
):  # pylint: disable=unused-argument
    """Get sample manifests project not owned case"""
    runner = CliRunner()
    mocker.patch.object(
        APIClient,
        "get_sample_manifests",
        side_effect=APIClientError(
            message="Project does not exist or you do not have privileges to access it",
            status_code=400,
        ),
    )
    with runner.isolated_filesystem():
        os.mkdir("test_dir")
        res = runner.invoke(
            get_sample_manifests,
            [
                str(uuid.uuid4()),
                "test_dir",
                *credentials,
            ],
        )
    assert res.exit_code == 0
    assert (
        "Project does not exist or you do not have privileges to access it"
        in res.output
    )


@assert_no_requests
def test_get_sample_manifests__bad_project_id(credentials, mocker):
    """Test manifest retrieve failure when non-uuid string is used as project
    id.
    """
    runner = CliRunner()
    mocked_get_sample_manifests = mocker.patch.object(
        APIClient,
        "get_sample_manifests",
    )
    with runner.isolated_filesystem():
        os.mkdir("test_dir")
        res = runner.invoke(
            get_sample_manifests,
            [
                "1111111",
                "test_dir",
                *credentials,
            ],
        )
    assert res.exit_code == 1
    mocked_get_sample_manifests.assert_not_called()
    assert "Project ID is not valid" in res.output


@assert_no_requests
def test_get_sample_manifests__bad_destination(credentials, mocker):
    """Test manifest retrieve failure when bad destination is supplied
    id.
    """
    runner = CliRunner()
    mocked_get_sample_manifests = mocker.patch.object(
        APIClient,
        "get_sample_manifests",
    )
    with runner.isolated_filesystem():
        res = runner.invoke(
            get_sample_manifests,
            [
                str(uuid.uuid4()),
                "bad_dir",
                *credentials,
            ],
        )
    assert res.exit_code == 1
    mocked_get_sample_manifests.assert_not_called()
    assert "is not a directory that exists" in res.output



"""Commands to be executed from command line."""
# pylint: disable=E0012,C0330,R0913
import click

from .get_sample_manifest.cli import get_sample_manifest


@click.group()
def sample_manifests():
    """Sample manifest management commands."""


sample_manifests.add_command(get_sample_manifest)



"""Samples command utilities."""
import json

import requests

from gencove import client  # noqa: I100
from gencove.logger import echo_debug, echo_warning  # noqa: I100
from gencove.utils import get_progress_bar

from .constants import CHUNK_SIZE


def fatal_process_sample_error(err):
    """Give up retrying if the error code is different from 403.

    If the error code is 403, we need to refresh the download url and try
    processing the sample again.

    Returns:
        bool: True if to giveup on backing off, False it to continue.
    """
    return err.response.status_code != 403


def download_file(destination, download_url, no_progress=False):
    """Download a file to file system.

    Args:
        destination (File): file object
        download_url (str): url of the file to download
        no_progress (bool): don't show progress bar

    Returns:
        str : file path
            location of the downloaded file
    """
    stream_params = dict(stream=True, allow_redirects=False, headers={}, timeout=30)

    with requests.get(download_url, **stream_params) as req:
        req.raise_for_status()
        echo_debug("Starting download")

        if not no_progress:
            pbar = get_progress_bar(int(req.headers["content-length"]), "Downloading: ")
            pbar.start()
        for chunk in req.iter_content(chunk_size=CHUNK_SIZE):
            destination.write(chunk)
            if not no_progress:
                pbar.update(pbar.value + len(chunk))
        if not no_progress:
            pbar.finish()

        echo_debug("Finished downloading")


def save_qc_file(destination, api_client, sample_id):
    """Helper function to save qc metrics to json file.

    Args:
        destination(BinaryIO): Buffer where qc metrics will be saved
        api_client(APIClient): an instance of the client to make a call
        sample_id(str): sample UUID as string

    Returns:
        None
    """
    try:
        sample_qcs = api_client.get_sample_qc_metrics(sample_id).results
    except client.APIClientError:
        echo_warning("Error getting sample quality control metrics.")
        raise
    content = json.dumps(sample_qcs, cls=client.CustomEncoder)
    writing_to_stdout = destination.isatty()
    if writing_to_stdout:
        content = content.encode()
    destination.write(content)


def save_metadata_file(destination, api_client, sample_id):
    """Helper function to save metadata to json file.

    Args:
        destination(BinaryIO): Buffer where qc metrics will be saved
        api_client(APIClient): an instance of the client to make a call
        sample_id(str): sample UUID as string

    Returns:
        None
    """
    try:
        metadata = api_client.get_metadata(sample_id)
    except client.APIClientError:
        echo_warning("Error getting sample metadata.")
        raise
    content = json.dumps(metadata, cls=client.CustomEncoder)
    writing_to_stdout = destination.isatty()
    if writing_to_stdout:
        content = content.encode()
    destination.write(content)



"""Project batches list shell command definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.command.utils import validate_uuid
from gencove.constants import Credentials, Optionals

from .main import ListBatches


@click.command("list-batches")
@click.argument("project_id", callback=validate_uuid)
@add_options(common_options)
def list_project_batches(project_id, host, email, password, api_key):
    """List batches that are available for a project.

    `PROJECT_ID`: Gencove project ID
    """
    ListBatches(
        project_id,
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
    ).run()



"""Delete projects shell command definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.command.utils import handle_exception, validate_uuid_list
from gencove.constants import Credentials, Optionals
from gencove.logger import echo_debug

from .main import Delete


@click.command("delete")
@click.argument(
    "project_ids",
    required=True,
    callback=validate_uuid_list,
)
@add_options(common_options)
def delete_projects(  # pylint: disable=too-many-arguments
    project_ids,
    host,
    email,
    password,
    api_key,
):
    """Delete projects.

    `PROJECT_IDS`: comma-separted list of projects to be deleted.
    """
    echo_debug(f"Project ids translation: {project_ids}")

    if not project_ids:
        handle_exception("No project ids provided")

    Delete(
        project_ids,
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
    ).run()



"""Project assign all uploads from a prefix shell command definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.command.utils import validate_uuid
from gencove.constants import Credentials, SampleAssignmentStatus
from gencove.utils import enum_as_dict


from .constants import RunPrefixOptionals
from .main import RunPrefix


@click.command("run-prefix")
@click.argument("project_id", callback=validate_uuid)
@click.argument("prefix")
@click.option(
    "--metadata-json",
    required=False,
    default=None,
    help=("Add metadata to all uploads that are to be assigned to a project."),
)
@click.option(
    "--status",
    help="Filter uploads by status of assignment",
    type=click.Choice(enum_as_dict(SampleAssignmentStatus).values()),
    default=SampleAssignmentStatus.ALL.value,
)
@add_options(common_options)
def run_prefix(  # pylint: disable=too-many-arguments
    project_id, prefix, metadata_json, status, host, email, password, api_key
):  # pylint: disable=C0301
    """Assign all uploads from Gencove prefix to a project. Optionally add
    metadata to the samples. Uploads can also be filtered through status.

    `PROJECT_ID`: Gencove project ID

    `RUN_PREFIX`: Gencove prefix where samples are located formated as gncv://path

    Examples:

        Assign uploads to a project:

            gencove projects run-prefix 06a5d04b-526a-4471-83ba-fb54e0941758 gncv://my-project/path

        Assign uploads to a project with metadata:

            gencove projects run-prefix 06a5d04b-526a-4471-83ba-fb54e0941758 gncv://my-project/path --metadata-json='{"batch": "batch1"}'

        Assign uploads filtered by status to a project:

            gencove projects run-prefix 06a5d04b-526a-4471-83ba-fb54e0941758 gncv://my-project/path --status assigned
    """  # noqa: E501
    RunPrefix(
        project_id,
        prefix,
        Credentials(email=email, password=password, api_key=api_key),
        RunPrefixOptionals(host=host, metadata_json=metadata_json, status=status),
    ).run()
